"""
Application wrappers
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'

# Imports
import os
import re
from pkg_resources import parse_version
from subprocess import check_output, STDOUT, CalledProcessError

# Presto and changeo imports
from presto.IO import printError, printWarning
from changeo.Defaults import default_igblastn_exec, default_igblastp_exec, default_tbl2asn_exec, \
                             default_igphyml_exec

# Defaults
default_igblast_output = 'legacy'


def runASN(fasta, template=None, exec=default_tbl2asn_exec):
    """
    Executes tbl2asn to generate Sequin files

    Arguments:
      fasta (str): fsa file name.
      template (str): sbt file name.
      exec (str): the name or path to the tbl2asn executable.

    Returns:
      str: tbl2asn console output.
    """
    # Basic command that requires .fsa and .tbl files in the same directory
    # tbl2asn -i records.fsa -a s -V vb -t template.sbt

    # Define tbl2asn command
    cmd = [exec,
           '-i', os.path.abspath(fasta),
           '-a', 's',
           '-V', 'vb']
    if template is not None:
        cmd.extend(['-t', os.path.abspath(template)])

    # Execute tbl2asn
    try:
        stdout_str = check_output(cmd, stderr=STDOUT, shell=False,
                                  universal_newlines=True)
    except CalledProcessError as e:
        printError('Running command: %s\n%s' % (' '.join(cmd), e.output))

    if 'Unable to read any FASTA records' in stdout_str:
        printError('%s failed: %s' % (' '.join(cmd), stdout_str))

    return stdout_str


def runIgPhyML(rep_file, rep_dir, model='HLP17', motifs='FCH',
               threads=1, exec=default_igphyml_exec):
    """
    Run IgPhyML

    Arguments:
      rep_file (str): repertoire tsv file.
      rep_dir (str): directory containing input fasta files.
      model (str): model to use.
      motif (str): motifs argument.
      threads : number of threads.
      exec : the path to the IgPhyMl executable.

    Returns:
      str: name of the output tree file.
    """
    # cd rep_dir
    # igphyml --repfile rep_file -m HLP17 --motifs FCH --omegaOpt e,e --run_id test -o tlr --threads 4 --minSeq 2

    # Define igphyml command
    cmd = [exec,
           '--repfile', rep_file,
           '-m', model,
           '--motifs', motifs,
           '--omegaOpt',  'e,e',
           '-o', 'tlr',
           '--minSeq', '2',
           '--threads', str(threads)]

    # Run IgPhyMl
    try:
        stdout_str = check_output(cmd, stderr=STDOUT, shell=False,
                                  universal_newlines=True, cwd=rep_dir)
    except CalledProcessError as e:
        printError('Running command: %s\n%s' % (' '.join(cmd), e.output))

    return None


def runIgBLASTN(fasta, igdata, loci='ig', organism='human', vdb=None, ddb=None, jdb=None, cdb=None,
                output=None, format=default_igblast_output, threads=1, exec=default_igblastn_exec):
    """
    Runs igblastn on a sequence file

    Arguments:
      fasta (str): fasta file containing sequences.
      igdata (str): path to the IgBLAST database directory (IGDATA environment).
      loci (str): receptor type; one of 'ig' or 'tr'.
      organism (str): species name.
      vdb (str): name of a custom V reference in the database folder to use.
      ddb (str): name of a custom D reference in the database folder to use.
      jdb (str): name of a custom J reference in the database folder to use.
      cdb (str): name of a custom C reference in the database folder to use.
      output (str): output file name. If None, automatically generate from the fasta file name.
      format (str): output format. One of 'blast' or 'airr'.
      threads (int): number of threads for igblastn.
      exec (str): the name or path to the igblastn executable.

    Returns:
      str: IgBLAST console output.

    """
    # export IGDATA
    # declare -A SEQTYPE
    # SEQTYPE[ig] = "Ig"
    # SEQTYPE[tr] = "TCR"
    # GERMLINE_V = "imgt_${SPECIES}_${RECEPTOR}_v"
    # GERMLINE_D = "imgt_${SPECIES}_${RECEPTOR}_d"
    # GERMLINE_J = "imgt_${SPECIES}_${RECEPTOR}_j"
    # GERMLINE_C = "imgt_${SPECIES}_${RECEPTOR}_c"
    # AUXILIARY = "${SPECIES}_gl.aux"
    # IGBLAST_DB = "${IGDATA}/database"
    # IGBLAST_CMD = "igblastn \
    #     -germline_db_V ${IGBLAST_DB}/${GERMLINE_V} \
    #     -germline_db_D ${IGBLAST_DB}/${GERMLINE_D} \
    #     -germline_db_J ${IGBLAST_DB}/${GERMLINE_J} \
    #     -c_region_db ${IGBLAST_DB}/${GERMLINE_C} \
    #     -auxiliary_data ${IGDATA}/optional_file/${AUXILIARY} \
    #     -ig_seqtype ${SEQTYPE[${RECEPTOR}]} -organism ${SPECIES} \
    #     -domain_system imgt -outfmt '7 std qseq sseq btop'"
    #
    # # Set run commmand
    # OUTFILE =$(basename ${READFILE})
    # OUTFILE = "${OUTDIR}/${OUTFILE%.fasta}.fmt7"
    # IGBLAST_VER =$(${IGBLAST_CMD} -version | grep 'Package' | sed s / 'Package: ' //)
    # IGBLAST_RUN = "${IGBLAST_CMD} -query ${READFILE} -out ${OUTFILE} -num_threads ${NPROC}"

    # Define arguments
    try:
        outfmt = {'blast': '7 std qseq sseq btop', 'airr': '19'}[format]
    except KeyError:
        printError('Invalid output format %s.' % format)

    try:
        seqtype = {'ig': 'Ig', 'tr': 'TCR'}[loci]
    except KeyError:
        printError('Invalid receptor type %s.' % loci)

    # Set auxiliary data
    auxiliary = os.path.join(igdata, 'optional_file', '%s_gl.aux' % organism)
    # Set V database
    if vdb is not None:  v_germ = os.path.join(igdata, 'database', vdb)
    else:  v_germ = os.path.join(igdata, 'database', 'imgt_%s_%s_v' % (organism, loci))
    # Set D database
    if ddb is not None:  d_germ = os.path.join(igdata, 'database', ddb)
    else:  d_germ = os.path.join(igdata, 'database', 'imgt_%s_%s_d' % (organism, loci))
    # Set J database
    if jdb is not None:  j_germ = os.path.join(igdata, 'database', jdb)
    else:  j_germ = os.path.join(igdata, 'database', 'imgt_%s_%s_j' % (organism, loci))
    # Set C database
    if cdb is not None:  c_germ = os.path.join(igdata, 'database', cdb)
    else:  c_germ = os.path.join(igdata, 'database', 'imgt_%s_%s_c' % (organism, loci))

    # Define IgBLAST command
    cmd = [exec,
           '-query', os.path.abspath(fasta),
           '-out', os.path.abspath(output),
           '-num_threads', str(threads),
           '-ig_seqtype', seqtype,
           '-organism', organism,
           '-auxiliary_data', str(auxiliary),
           '-germline_db_V', str(v_germ),
           '-germline_db_D', str(d_germ),
           '-germline_db_J', str(j_germ),
           '-outfmt', outfmt,
           '-domain_system', 'imgt']

    # Add C-region arguments for igblastn v1.18.0
    version = getIgBLASTVersion(exec=exec)
    if parse_version(version) >= parse_version('1.18.0'):
        cmd = cmd + ['-c_region_db', str(c_germ)]
    #print(cmd)

    # Execute IgBLAST
    env = os.environ.copy()
    env['IGDATA'] = igdata
    try:
        stdout_str = check_output(cmd, stderr=STDOUT, shell=False, env=env,
                                  universal_newlines=True)
    except CalledProcessError as e:
        printError('Running command: %s\n%s' % (' '.join(cmd), e.output))

    #if 'Unable to read any FASTA records' in stdout_str:
    #    sys.stderr.write('\n%s failed: %s\n' % (' '.join(cmd), stdout_str))

    return stdout_str


def runIgBLASTP(fasta, igdata, loci='ig', organism='human', vdb=None, output=None,
                threads=1, exec=default_igblastp_exec):
    """
    Runs igblastp on a sequence file

    Arguments:
      fasta (str): fasta file containing sequences.
      igdata (str): path to the IgBLAST database directory (IGDATA environment).
      loci (str): receptor type; one of 'ig' or 'tr'.
      organism (str): species name.
      vdb (str): name of a custom V reference in the database folder to use.
      output (str): output file name. If None, automatically generate from the fasta file name.
      threads (int): number of threads for igblastp.
      exec (str): the name or path to the igblastp executable.

    Returns:
      str: IgBLAST console output.

    """
    # IGBLAST_CMD="igblastp \
    #     -germline_db_V ${IGBLAST_DB}/imgt_aa_${SPECIES}_${RECEPTOR}_v \
    #     -ig_seqtype ${SEQTYPE} -organism ${SPECIES} \
    #     -domain_system imgt -outfmt '7 std qseq sseq btop'"

    try:
        seqtype = {'ig': 'Ig', 'tr': 'TCR'}[loci]
    except KeyError:
        printError('Invalid receptor type %s.' % loci)

    # Set V database
    if vdb is not None:  v_germ = os.path.join(igdata, 'database', vdb)
    else:  v_germ = os.path.join(igdata, 'database', 'imgt_aa_%s_%s_v' % (organism, loci))

    # Define IgBLAST command
    cmd = [exec,
           '-query', os.path.abspath(fasta),
           '-out', os.path.abspath(output),
           '-num_threads', str(threads),
           '-ig_seqtype', seqtype,
           '-organism', organism,
           '-germline_db_V', str(v_germ),
           '-outfmt', '7 std qseq sseq btop',
           '-domain_system', 'imgt']

    # Execute IgBLAST
    env = os.environ.copy()
    env['IGDATA'] = igdata
    try:
        stdout_str = check_output(cmd, stderr=STDOUT, shell=False, env=env,
                                  universal_newlines=True)
    except CalledProcessError as e:
        printError('Running command: %s\n%s' % (' '.join(cmd), e.output))

    return stdout_str


def getIgBLASTVersion(exec=default_igblastn_exec):
    """
    Gets the version of the IgBLAST executable

    Arguments:
      exec (str): the name or path to the igblastn executable.

    Returns:
      str: version number.
    """
    # Build commandline
    cmd = [exec, '-version']

    # Run
    try:
        stdout_str = check_output(cmd, stderr=STDOUT, shell=False, universal_newlines=True)
    except CalledProcessError as e:
        printError('Running command: %s\n%s' % (' '.join(cmd), e.output))

    # Extract version number
    match = re.search(r'(?<=Package: igblast )(\d+\.\d+\.\d+)', stdout_str)
    version = match.group(0)

    return version


"""
Unit tests for MakeDb
"""

# Imports
import os
import time
import unittest

# Presto and changeo imports
from changeo.Receptor import Receptor
from changeo.Alignment import decodeBTOP, decodeCIGAR, encodeCIGAR, padAlignment, alignmentPositions

# Paths
test_path = os.path.dirname(os.path.realpath(__file__))
data_path = os.path.join(test_path, 'data')


class Test_Receptor(unittest.TestCase):
    def setUp(self):
        print('-> %s()' % self._testMethodName)

        row = {'SEQUENCE_ID':    'TEST1',
               'SEQUENCE_INPUT': 'TGTGCAAGGGGGCCATTGGACTACTTCTACTACGGTGTGGACGNNNNN',
               'V_CALL':         'IGHV6-1*01,IGHV6-1*02',
               'D_CALL':         'IGHD6-6*01',
               'J_CALL':         'IGHJ6*02',}
        self.ig_rec = Receptor(row)

        row = {'SEQUENCE_ID':    'TEST2',
               'SEQUENCE_INPUT': 'TGTGCAAGGGGGCCATTGGACTACTTCTACTACGGTGTGGACGNNNNN',
               'V_CALL':         'TRAV6-1*01,TRAV6-1*02',
               'D_CALL':         'TRAD6-6*01',
               'J_CALL':         'TRAJ6*02',}
        self.tr_rec = Receptor(row)

        row = {'SEQUENCE_ID':    'TEST3',
               'SEQUENCE_INPUT': '',
               'V_CALL':         '',
               'D_CALL':         None,
               'J_CALL':         'princess,banana,hammock',}
        self.bad_rec = Receptor(row)

        # CIGAR strings
        self.cigar_string = ['30M1I69M3D', '4S16N30M1I69M3D']
        self.cigar_decoded = [[('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                              [('S', 4), ('N', 16), ('M', 30), ('I', 1), ('M', 69), ('D', 3)]]
        self.cigar_len = [{'q_start': 0, 'q_length': 30 + 1 + 69 + 0, 'r_start': 0, 'r_length': 30 + 0 + 69 + 3},
                          {'q_start': 4, 'q_length': 30 + 1 + 69 + 0, 'r_start': 16, 'r_length': 30 + 0 + 69 + 3}]

        self.cigar_pos = [(0, 0), (5, 0), (0, 3), (5, 3)]
        self.cigar_pad_1 = [[('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                            [('S', 5), ('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                            [('N', 3), ('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                            [('S', 5), ('N', 3), ('M', 30), ('I', 1), ('M', 69), ('D', 3)]]
        self.cigar_pad_2 = [[('S', 4), ('N', 16), ('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                            [('S', 9), ('N', 16), ('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                            [('S', 4), ('N', 19), ('M', 30), ('I', 1), ('M', 69), ('D', 3)],
                            [('S', 9), ('N', 19), ('M', 30), ('I', 1), ('M', 69), ('D', 3)]]

        # BTOP strings
        self.btop_string = ['7AGAC39',
                            '7A-39',
                            '6-G-A41',
                            'AG8-GC-CTCT']
        self.btop_decoded_full = [[('=', 7), ('X', 2), ('=', 39)],
                                  [('=', 7), ('D', 1), ('=', 39)],
                                  [('=', 6), ('I', 2), ('=', 41)],
                                  [('X', 1), ('=', 8), ('I', 1), ('D', 1), ('X', 2)]]

        self.start = time.time()

    def tearDown(self):
        t = time.time() - self.start
        print('<- %s() %.3f' % (self._testMethodName, t))

    @unittest.skip("-> Receptor() skipped\n")
    def test_Receptor(self):
        print('IG>')
        print(self.ig_rec.getAlleleCalls(['v','d','j'], action='first'))
        print(self.ig_rec.getAlleleCalls(['v','j'], action='first'))
        print(self.ig_rec.getAlleleCalls(['j','v'], action='first'))
        print(self.ig_rec.getGeneCalls(['v','d','j'], action='first'))
        print(self.ig_rec.getGeneCalls(['v','j'], action='first'))
        print(self.ig_rec.getGeneCalls(['j','v'], action='first'))
        print(self.ig_rec.getFamilyCalls(['v','d','j'], action='first'))
        print(self.ig_rec.getFamilyCalls(['v','j'], action='first'))
        print(self.ig_rec.getFamilyCalls(['j','v'], action='first'))

        print(self.ig_rec.getAlleleCalls(['v','d','j'], action='set'))
        print(self.ig_rec.getAlleleCalls(['v','j'], action='set'))
        print(self.ig_rec.getAlleleCalls(['j','v'], action='set'))
        print(self.ig_rec.getGeneCalls(['v','d','j'], action='set'))
        print(self.ig_rec.getGeneCalls(['v','j'], action='set'))
        print(self.ig_rec.getGeneCalls(['j','v'], action='set'))
        print(self.ig_rec.getFamilyCalls(['v','d','j'], action='set'))
        print(self.ig_rec.getFamilyCalls(['v','j'], action='set'))
        print(self.ig_rec.getFamilyCalls(['j','v'], action='set'))

        print(self.ig_rec.getAlleleCalls(['v','d','j'], action='list'))
        print(self.ig_rec.getAlleleCalls(['v','j'], action='list'))
        print(self.ig_rec.getAlleleCalls(['j','v'], action='list'))
        print(self.ig_rec.getGeneCalls(['v','d','j'], action='list'))
        print(self.ig_rec.getGeneCalls(['v','j'], action='list'))
        print(self.ig_rec.getGeneCalls(['j','v'], action='list'))
        print(self.ig_rec.getFamilyCalls(['v','d','j'], action='list'))
        print(self.ig_rec.getFamilyCalls(['v','j'], action='list'))
        print(self.ig_rec.getFamilyCalls(['j','v'], action='list'))

        print('TR>')
        print(self.tr_rec.getAlleleCalls(['v','d','j'], action='first'))
        print(self.tr_rec.getAlleleCalls(['v','j'], action='first'))
        print(self.tr_rec.getAlleleCalls(['j','v'], action='first'))
        print(self.tr_rec.getGeneCalls(['v','d','j'], action='first'))
        print(self.tr_rec.getGeneCalls(['v','j'], action='first'))
        print(self.tr_rec.getGeneCalls(['j','v'], action='first'))
        print(self.tr_rec.getFamilyCalls(['v','d','j'], action='first'))
        print(self.tr_rec.getFamilyCalls(['v','j'], action='first'))
        print(self.tr_rec.getFamilyCalls(['j','v'], action='first'))

        print(self.tr_rec.getAlleleCalls(['v','d','j'], action='set'))
        print(self.tr_rec.getAlleleCalls(['v','j'], action='set'))
        print(self.tr_rec.getAlleleCalls(['j','v'], action='set'))
        print(self.tr_rec.getGeneCalls(['v','d','j'], action='set'))
        print(self.tr_rec.getGeneCalls(['v','j'], action='set'))
        print(self.tr_rec.getGeneCalls(['j','v'], action='set'))
        print(self.tr_rec.getFamilyCalls(['v','d','j'], action='set'))
        print(self.tr_rec.getFamilyCalls(['v','j'], action='set'))
        print(self.tr_rec.getFamilyCalls(['j','v'], action='set'))

        print(self.tr_rec.getAlleleCalls(['v','d','j'], action='list'))
        print(self.tr_rec.getAlleleCalls(['v','j'], action='list'))
        print(self.tr_rec.getAlleleCalls(['j','v'], action='list'))
        print(self.tr_rec.getGeneCalls(['v','d','j'], action='list'))
        print(self.tr_rec.getGeneCalls(['v','j'], action='list'))
        print(self.tr_rec.getGeneCalls(['j','v'], action='list'))
        print(self.tr_rec.getFamilyCalls(['v','d','j'], action='list'))
        print(self.tr_rec.getFamilyCalls(['v','j'], action='list'))
        print(self.tr_rec.getFamilyCalls(['j','v'], action='list'))

        print('JUNK>')
        print(self.bad_rec.getAlleleCalls(['v','d','j'], action='first'))
        print(self.bad_rec.getAlleleCalls(['v','j'], action='first'))
        print(self.bad_rec.getAlleleCalls(['j','v'], action='first'))
        print(self.bad_rec.getGeneCalls(['v','d','j'], action='first'))
        print(self.bad_rec.getGeneCalls(['v','j'], action='first'))
        print(self.bad_rec.getGeneCalls(['j','v'], action='first'))
        print(self.bad_rec.getFamilyCalls(['v','d','j'], action='first'))
        print(self.bad_rec.getFamilyCalls(['v','j'], action='first'))
        print(self.bad_rec.getFamilyCalls(['j','v'], action='first'))

        print(self.bad_rec.getAlleleCalls(['v','d','j'], action='set'))
        print(self.bad_rec.getAlleleCalls(['v','j'], action='set'))
        print(self.bad_rec.getAlleleCalls(['j','v'], action='set'))
        print(self.bad_rec.getGeneCalls(['v','d','j'], action='set'))
        print(self.bad_rec.getGeneCalls(['v','j'], action='set'))
        print(self.bad_rec.getGeneCalls(['j','v'], action='set'))
        print(self.bad_rec.getFamilyCalls(['v','d','j'], action='set'))
        print(self.bad_rec.getFamilyCalls(['v','j'], action='set'))
        print(self.bad_rec.getFamilyCalls(['j','v'], action='set'))

        print(self.bad_rec.getAlleleCalls(['v','d','j'], action='list'))
        print(self.bad_rec.getAlleleCalls(['v','j'], action='list'))
        print(self.bad_rec.getAlleleCalls(['j','v'], action='list'))
        print(self.bad_rec.getGeneCalls(['v','d','j'], action='list'))
        print(self.bad_rec.getGeneCalls(['v','j'], action='list'))
        print(self.bad_rec.getGeneCalls(['j','v'], action='list'))
        print(self.bad_rec.getFamilyCalls(['v','d','j'], action='list'))
        print(self.bad_rec.getFamilyCalls(['v','j'], action='list'))
        print(self.bad_rec.getFamilyCalls(['j','v'], action='list'))

        self.fail('TODO')

    #@unittest.skip("-> decodeCIGAR() skipped\n")
    def test_decodeCIGAR(self):
        for cigar, truth in zip(self.cigar_string, self.cigar_decoded):
            result = decodeCIGAR(cigar)
            print(result)
            self.assertListEqual(truth, result)

    #@unittest.skip("-> decodeBTOP() skipped\n")
    def test_decodeBTOP(self):
        for btop, truth in zip(self.btop_string, self.btop_decoded_full):
            result = decodeBTOP(btop)
            print('FULL> ', result)
            self.assertListEqual(truth, result)

    #@unittest.skip("-> encodeCIGAR() skipped\n")
    def test_encodeCIGAR(self):
        for align, truth in zip(self.cigar_decoded, self.cigar_string):
            result = encodeCIGAR(align)
            print(result)
            self.assertEqual(truth, result)

    #@unittest.skip("-> padAlignment() skipped\n")
    def test_padAlignment(self):
        cigar = self.cigar_decoded[0]
        for (s, r), truth in zip(self.cigar_pos, self.cigar_pad_1):
            #print('POS>', s, r)
            result = padAlignment(cigar, s, r)
            print('PAD>', '(%i, %i) =' % (s, r), result)
            self.assertEqual(truth, result)

        cigar = self.cigar_decoded[1]
        for (s, r), truth in zip(self.cigar_pos, self.cigar_pad_2):
            #print('POS>', s, r)
            result = padAlignment(cigar, s, r)
            print('PAD>', '(%i, %i) =' % (s, r), result)
            self.assertEqual(truth, result)

    #@unittest.skip("-> alignmentPositions() skipped\n")
    def test_alignmentPositions(self):
        for align, truth in zip(self.cigar_decoded, self.cigar_len):
            result = alignmentPositions(align)
            print('POS>', result)
            self.assertDictEqual(truth, result)


if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
"""
Parses tab delimited database files
"""

# Info
__author__ = 'Jason Anthony Vander Heiden'
from changeo import __version__, __date__

# Imports
import csv
import os
import re
import shutil
from argparse import ArgumentParser
from collections import OrderedDict
from itertools import chain
from textwrap import dedent
from time import time
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

# Presto and changeo imports
from presto.Annotation import flattenAnnotation
from presto.IO import printLog, printMessage, printProgress, printError, printWarning
from changeo.Applications import default_tbl2asn_exec, runASN
from changeo.Defaults import default_id_field, default_seq_field, default_germ_field, \
                             default_csv_size, default_format, default_out_args
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs
from changeo.Gene import getCGene
from changeo.IO import countDbFile, getFormatOperators, getOutputHandle, AIRRReader, AIRRWriter, \
                       ChangeoReader, ChangeoWriter, TSVReader, ReceptorData, readGermlines, \
                       checkFields, yamlDict
from changeo.Receptor import AIRRSchema, ChangeoSchema

# System settings
csv.field_size_limit(default_csv_size)

# Defaults
default_db_xref = 'IMGT/GENE-DB'
default_molecule = 'mRNA'
default_product = 'immunoglobulin heavy chain'
default_allele_delim = '*'


def buildSeqRecord(db_record, id_field, seq_field, meta_fields=None):
    """
    Parses a database record into a SeqRecord

    Arguments: 
      db_record : a dictionary containing a database record.
      id_field : the field containing identifiers.
      seq_field : the field containing sequences.
      meta_fields : a list of fields to add to sequence annotations.

    Returns: 
      Bio.SeqRecord.SeqRecord: record.
    """
    # Return None if ID or sequence fields are empty
    if not db_record[id_field] or not db_record[seq_field]:
        return None
    
    # Create description string
    desc_dict = OrderedDict([('ID', db_record[id_field])])
    if meta_fields is not None:
        desc_dict.update([(f, db_record[f]) for f in meta_fields if f in db_record]) 
    desc_str = flattenAnnotation(desc_dict)
    
    # Create SeqRecord
    seq_record = SeqRecord(Seq(db_record[seq_field]),
                           id=desc_str, name=desc_str, description='')
        
    return seq_record


def convertToAIRR(db_file, format=default_format,
                  out_file=None, out_args=default_out_args):
    """
    Converts a Change-O formatted file into an AIRR formatted file

    Arguments:
      db_file : the database file name.
      format : input format.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
     str : output file name
    """
    log = OrderedDict()
    log['START'] = 'ConvertDb'
    log['COMMAND'] = 'airr'
    log['FILE'] = os.path.basename(db_file)
    printLog(log)

    # Define format operators
    try:
        reader, __, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)

    # Open input
    db_handle = open(db_file, 'rt')
    db_iter = reader(db_handle)

    # Set output fields replacing length with end fields
    in_fields = [schema.toReceptor(f) for f in db_iter.fields]
    out_fields = []
    for f in in_fields:
        if f in ReceptorData.length_fields and ReceptorData.length_fields[f][0] in in_fields:
            out_fields.append(ReceptorData.length_fields[f][1])
        out_fields.append(f)
    out_fields = list(OrderedDict.fromkeys(out_fields))
    out_fields = [AIRRSchema.fromReceptor(f) for f in out_fields]

    # Open output writer
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(db_file, out_label='airr', out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'], out_type=AIRRSchema.out_type)
    pass_writer = AIRRWriter(pass_handle, fields=out_fields)

    # Count records
    result_count = countDbFile(db_file)

    # Iterate over records
    start_time = time()
    rec_count = 0
    for rec in db_iter:
        # Print progress for previous iteration
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += 1
        # Write records
        pass_writer.writeReceptor(rec)

    # Print counts
    printProgress(rec_count, result_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['RECORDS'] = rec_count
    log['END'] = 'ConvertDb'
    printLog(log)

    # Close file handles
    pass_handle.close()
    db_handle.close()

    return pass_handle.name


def convertToChangeo(db_file, out_file=None, out_args=default_out_args):
    """
    Converts an AIRR formatted file into an Change-O formatted file

    Arguments:
      db_file: the database file name.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
      str : output file name.
    """
    log = OrderedDict()
    log['START'] = 'ConvertDb'
    log['COMMAND'] = 'changeo'
    log['FILE'] = os.path.basename(db_file)
    printLog(log)

    # Open input
    db_handle = open(db_file, 'rt')
    db_iter = AIRRReader(db_handle)

    # Set output fields replacing length with end fields
    in_fields = [AIRRSchema.toReceptor(f) for f in db_iter.fields]
    out_fields = []
    for f in in_fields:
        out_fields.append(f)
        if f in ReceptorData.end_fields and ReceptorData.end_fields[f][0] in in_fields:
            out_fields.append(ReceptorData.end_fields[f][1])
    out_fields = list(OrderedDict.fromkeys(out_fields))
    out_fields = [ChangeoSchema.fromReceptor(f) for f in out_fields]

    # Open output writer
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(db_file, out_label='changeo', out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'], out_type=ChangeoSchema.out_type)
    pass_writer = ChangeoWriter(pass_handle, fields=out_fields)

    # Count records
    result_count = countDbFile(db_file)

    # Iterate over records
    start_time = time()
    rec_count = 0
    for rec in db_iter:
        # Print progress for previous iteration
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += 1
        # Write records
        pass_writer.writeReceptor(rec)

    # Print counts
    printProgress(rec_count, result_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['RECORDS'] = rec_count
    log['END'] = 'ConvertDb'
    printLog(log)

    # Close file handles
    pass_handle.close()
    db_handle.close()

    return pass_handle.name


# TODO:  SHOULD ALLOW FOR UNSORTED CLUSTER COLUMN
# TODO:  SHOULD ALLOW FOR GROUPING FIELDS
def convertToBaseline(db_file, id_field=default_id_field, seq_field=default_seq_field,
                      germ_field=default_germ_field, cluster_field=None,
                      meta_fields=None, out_file=None, out_args=default_out_args):
    """
    Builds fasta files from database records

    Arguments: 
      db_file : the database file name.
      id_field : the field containing identifiers.
      seq_field : the field containing sample sequences.
      germ_field : the field containing germline sequences.
      cluster_field : the field containing clonal groupings;
                    if None write the germline for each record.
      meta_fields : a list of fields to add to sequence annotations.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.
                    
    Returns: 
     str : output file name
    """
    log = OrderedDict()
    log['START'] = 'ConvertDb'
    log['COMMAND'] = 'fasta'
    log['FILE'] = os.path.basename(db_file)
    log['ID_FIELD'] = id_field
    log['SEQ_FIELD'] = seq_field
    log['GERM_FIELD'] = germ_field
    log['CLUSTER_FIELD'] = cluster_field
    if meta_fields is not None:  log['META_FIELDS'] = ','.join(meta_fields)
    printLog(log)
    
    # Open input
    db_handle = open(db_file, 'rt')
    db_iter = TSVReader(db_handle)
    result_count = countDbFile(db_file)

    # Open output
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(db_file, out_label='sequences', out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'], out_type='clip')
    # Iterate over records
    start_time = time()
    rec_count, germ_count, pass_count, fail_count = 0, 0, 0, 0
    cluster_last = None
    for rec in db_iter:
        # Print progress for previous iteration
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += 1
        
        # Update cluster ID
        cluster = rec.get(cluster_field, None)
        
        # Get germline SeqRecord when needed
        if cluster_field is None:
            germ = buildSeqRecord(rec, id_field, germ_field, meta_fields)
            germ.id = '>' + germ.id
        elif cluster != cluster_last:
            germ = buildSeqRecord(rec, cluster_field, germ_field)
            germ.id = '>' + germ.id            
        else:
            germ = None

        # Get read SeqRecord
        seq = buildSeqRecord(rec, id_field, seq_field, meta_fields)
        
        # Write germline
        if germ is not None:
            germ_count += 1
            SeqIO.write(germ, pass_handle, 'fasta')
        
        # Write sequences
        if seq is not None:
            pass_count += 1
            SeqIO.write(seq, pass_handle, 'fasta')
        else:
            fail_count += 1
        
        # Set last cluster ID
        cluster_last = cluster
        
    # Print counts
    printProgress(rec_count, result_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['RECORDS'] = rec_count
    log['GERMLINES'] = germ_count
    log['PASS'] = pass_count
    log['FAIL'] = fail_count
    log['END'] = 'ConvertDb'
    printLog(log)

    # Close file handles
    pass_handle.close()
    db_handle.close()

    return pass_handle.name


def convertToFasta(db_file, id_field=default_id_field, seq_field=default_seq_field,
                   meta_fields=None, out_file=None, out_args=default_out_args):
    """
    Builds fasta files from database records

    Arguments: 
      db_file : the database file name.
      id_field : the field containing identifiers.
      seq_field : the field containing sequences.
      meta_fields : a list of fields to add to sequence annotations.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.
                    
    Returns: 
      str : output file name.
    """
    log = OrderedDict()
    log['START'] = 'ConvertDb'
    log['COMMAND'] = 'fasta'
    log['FILE'] = os.path.basename(db_file)
    log['ID_FIELD'] = id_field
    log['SEQ_FIELD'] = seq_field
    if meta_fields is not None:  log['META_FIELDS'] = ','.join(meta_fields)
    printLog(log)
    
    # Open input
    out_type = 'fasta'
    db_handle = open(db_file, 'rt')
    db_iter = TSVReader(db_handle)
    result_count = countDbFile(db_file)

    # Open output
    if out_file is not None:
        pass_handle = open(out_file, 'w')
    else:
        pass_handle = getOutputHandle(db_file, out_label='sequences', out_dir=out_args['out_dir'],
                                      out_name=out_args['out_name'], out_type=out_type)

    # Iterate over records
    start_time = time()
    rec_count, pass_count, fail_count = 0, 0, 0
    for rec in db_iter:
        # Print progress for previous iteration
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += 1

        # Get SeqRecord
        seq = buildSeqRecord(rec, id_field, seq_field, meta_fields)

        # Write sequences
        if seq is not None:
            pass_count += 1
            SeqIO.write(seq, pass_handle, out_type)
        else:
            fail_count += 1
        
    # Print counts
    printProgress(rec_count, result_count, 0.05, start_time=start_time)
    log = OrderedDict()
    log['OUTPUT'] = os.path.basename(pass_handle.name)
    log['RECORDS'] = rec_count
    log['PASS'] = pass_count
    log['FAIL'] = fail_count
    log['END'] = 'ConvertDb'
    printLog(log)

    # Close file handles
    pass_handle.close()
    db_handle.close()

    return pass_handle.name


def makeGenbankFeatures(record, start=None, end=None, product=default_product,
                        inference=None, db_xref=None,
                        c_field=None, allow_stop=False, asis_calls=False,
                        allele_delim=default_allele_delim):
    """
    Creates a feature table for GenBank submissions

    Arguments:
      record : Receptor record.
      start : start position of the modified sequence in the input sequence. Used for feature position offsets.
      end : end position of the modified sequence in the input sequence. Used for feature position offsets.
      product : Product (protein) name.
      inference : Reference alignment tool.
      db_xref : Reference database name.
      c_field : column containing the C region gene call.
      allow_stop : if True retain records with junctions having stop codons.
      asis_calls : if True do not parse gene calls for IMGT nomenclature.
      allele_delim : delimiter separating the gene name from the allele number when asis_calls=True.

    Returns:
      dict : dictionary defining GenBank features where the key is a tuple
             (start, end, feature key) and values are a list of
             tuples contain (qualifier key, qualifier value).
    """
    # .tbl file format
    #   Line 1, Column 1: Start location of feature
    #   Line 1, Column 2: Stop location of feature
    #   Line 1, Column 3: Feature key
    #   Line 2, Column 4: Qualifier key
    #   Line 2, Column 5: Qualifier value

    # Get genes and alleles
    c_gene = None
    if not asis_calls:
        # V gene
        v_gene = record.getVGene()
        v_allele = record.getVAlleleNumber()
        # D gene
        d_gene = record.getDGene()
        d_allele = record.getDAlleleNumber()
        # J gene
        j_gene = record.getJGene()
        j_allele = record.getJAlleleNumber()
        # C region
        if c_field is not None:
            c_gene = getCGene(record.getField(c_field), action='first')
    else:
        # V gene
        v_split = iter(record.v_call.rsplit(allele_delim, maxsplit=1))
        v_gene = next(v_split, None)
        v_allele = next(v_split, None)
        # D gene
        d_split = iter(record.d_call.rsplit(allele_delim, maxsplit=1))
        d_gene = next(d_split, None)
        d_allele = next(d_split, None)
        # J gene
        j_split = iter(record.j_call.rsplit(allele_delim, maxsplit=1))
        j_gene = next(j_split, None)
        j_allele = next(j_split, None)
        # C region
        if c_field is not None:
            c_gene = record.getField(c_field)

    # Fail if V or J is missing
    if v_gene is None or j_gene is None:
        return None

    # Set position offsets if required
    start_trim = 0 if start is None else start
    end_trim = 0 if end is None else len(record.sequence_input) - end
    source_len = len(record.sequence_input) - end_trim

    # Define return object
    result = OrderedDict()

    # C_region
    #   gene
    #   db_xref
    #   inference
    c_region_start = record.j_seq_end + 1 - start_trim
    c_region_length = len(record.sequence_input[(c_region_start + start_trim - 1):]) - end_trim
    if c_region_length > 0:
        if c_gene is not None:
            c_region = [('gene', c_gene)]
            if db_xref is not None:
                c_region.append(('db_xref', '%s:%s' % (db_xref, c_gene)))
        else:
            c_region = []

        # Assign C_region feature
        c_region_end = c_region_start + c_region_length - 1
        result[(c_region_start, '>%i' % c_region_end, 'C_region')] = c_region

        # Preserve J segment end position
        j_end = record.j_seq_end

        # Check for range error
        if c_region_end > source_len:
            return None
    else:
        # Trim J segment end position
        j_end = record.j_seq_end + c_region_length

    # V_region
    variable_start = max(record.v_seq_start - start_trim, 1)
    variable_end = j_end - start_trim
    result[(variable_start, variable_end, 'V_region')] = []

    # Check for range error
    if variable_end > source_len:
        return None

    # Product feature
    result[(variable_start, variable_end, 'misc_feature')] = [('note', '%s variable region' % product)]

    # V_segment
    #   gene (gene name)
    #   allele (allele only, without gene name, don't use if ambiguous)
    #   db_xref (database link)
    #   inference (reference alignment tool)
    v_segment = [('gene', v_gene)]
    if v_allele is not None:
        v_segment.append(('allele', v_allele))
    if db_xref is not None:
        v_segment.append(('db_xref', '%s:%s' % (db_xref, v_gene)))
    if inference is not None:
        v_segment.append(('inference', 'COORDINATES:alignment:%s' % inference))
    result[(variable_start, record.v_seq_end - start_trim, 'V_segment')] = v_segment

    # D_segment
    #   gene
    #   allele
    #   db_xref
    #   inference
    if d_gene:
        d_segment = [('gene', d_gene)]
        if d_allele is not None:
            d_segment.append(('allele', d_allele))
        if db_xref is not None:
            d_segment.append(('db_xref', '%s:%s' % (db_xref, d_gene)))
        if inference is not None:
            d_segment.append(('inference', 'COORDINATES:alignment:%s' % inference))
        result[(record.d_seq_start - start_trim, record.d_seq_end - start_trim, 'D_segment')] = d_segment

    # J_segment
    #   gene
    #   allele
    #   db_xref
    #   inference
    j_segment = [('gene', j_gene)]
    if j_allele is not None:
        j_segment.append(('allele', j_allele))
    if db_xref is not None:
        j_segment.append(('db_xref', '%s:%s' % (db_xref, j_gene)))
    if inference is not None:
        j_segment.append(('inference', 'COORDINATES:alignment:%s' % inference))
    result[(record.j_seq_start - start_trim, j_end - start_trim, 'J_segment')] = j_segment

    # CDS
    #   codon_start (must indicate codon offset)
    #   function = JUNCTION
    #   inference
    # print(record.v_germ_end_imgt, record.v_seq_end, record.v_germ_end_imgt - 310)
    # print(record.junction_start, record.junction_end, record.junction_length)
    if record.junction_start is not None and record.junction_end is not None:
        # Define junction boundaries
        junction_start = record.junction_start - start_trim
        junction_end = record.junction_end - start_trim

        # CDS record
        cds_start = '<%i' % junction_start
        cds_end = '>%i' % junction_end
        cds_record = [('function', 'JUNCTION')]
        if inference is not None:
            cds_record.append(('inference', 'COORDINATES:protein motif:%s' % inference))

        # Check for valid translation
        junction_seq = record.sequence_input[(junction_start - 1):junction_end]
        if len(junction_seq) % 3 > 0:  junction_seq = junction_seq + 'N' * (3 - len(junction_seq) % 3)
        junction_aa = Seq(junction_seq).translate()

        # Return invalid record upon junction stop codon
        if '*' in junction_aa and not allow_stop:
            return None
        elif '*' in junction_aa:
            cds_record.append(('note', '%s junction region' % product))
            result[(cds_start, cds_end, 'misc_feature')] = cds_record
        else:
            cds_record.append(('product', '%s junction region' % product))
            cds_record.append(('codon_start', 1))
            result[(cds_start, cds_end, 'CDS')] = cds_record

    return result


def makeGenbankSequence(record, name=None, label=None, count_field=None, index_field=None,
                        molecule=default_molecule, features=None):
    """
    Creates a sequence for GenBank submissions

    Arguments:
      record : Receptor record.
      name : sequence identifier for the output sequence. If None,
             use the original sequence identifier.
      label : a string to use as a label for the ID. if None do not add a field label.
      count_field : field name to populate the AIRR_READ_COUNT note.
      index_field : field name to populate the AIRR_CELL_INDEX note.
      molecule : source molecule (eg, "mRNA", "genomic DNA")
      features : dictionary of sample features (BioSample attributes) to add to the description of each record.

    Returns:
      dict: dictionary with {'record': SeqRecord,
                             'start': start position in raw sequence,
                             'end': end position in raw sequence}
    """
    # Replace gaps with N
    seq = record.sequence_input
    seq = seq.replace('-', 'N').replace('.', 'N')

    # Strip leading and trailing Ns
    head_match = re.search('^N+', seq)
    tail_match = re.search('N+$', seq)
    seq_start = head_match.end() if head_match else 0
    seq_end = tail_match.start() if tail_match else len(seq)

    # Define ID
    if name is None:
        name = record.sequence_id.split(' ')[0]
    if label is not None:
        name = '%s=%s' % (label, name)
    if features is not None:
        sample_desc = ' '.join(['[%s=%s]' % (k, v) for k, v in features.items()])
        name = '%s %s' % (name, sample_desc)
    name = '%s [moltype=%s] [keyword=TLS; Targeted Locus Study; AIRR; MiAIRR:1.0]' % (name, molecule)

    # Notes
    note_dict = OrderedDict()
    if count_field is not None:
        note_dict['AIRR_READ_COUNT'] = record.getField(count_field)
    if index_field is not None:
        note_dict['AIRR_CELL_INDEX'] = record.getField(index_field)
    if note_dict:
        note = '; '.join(['%s:%s' % (k, v) for k, v in note_dict.items()])
        name = '%s [note=%s]' % (name, note)

    # Return SeqRecord and positions
    record = SeqRecord(Seq(seq[seq_start:seq_end]), id=name, name=name, description='')
    result = {'record': record, 'start': seq_start, 'end': seq_end}

    return result


def convertToGenbank(db_file, inference=None, db_xref=None, molecule=default_molecule,
                     product=default_product, features=None, c_field=None, label=None,
                     count_field=None, index_field=None, allow_stop=False,
                     asis_id=False, asis_calls=False, allele_delim=default_allele_delim,
                     build_asn=False, asn_template=None, tbl2asn_exec=default_tbl2asn_exec,
                     format=default_format, out_file=None,
                     out_args=default_out_args):
    """
    Builds GenBank submission fasta and table files

    Arguments:
      db_file : the database file name.
      inference : reference alignment tool.
      db_xref : reference database link.
      molecule : source molecule (eg, "mRNA", "genomic DNA")
      product : Product (protein) name.
      features : dictionary of sample features (BioSample attributes) to add to the description of each record.
      c_field : column containing the C region gene call.
      label : a string to use as a label for the ID. if None do not add a field label.
      count_field : field name to populate the AIRR_READ_COUNT note.
      index_field : field name to populate the AIRR_CELL_INDEX note.
      allow_stop : if True retain records with junctions having stop codons.
      asis_id : if True use the original sequence ID for the output IDs.
      asis_calls : if True do not parse gene calls for IMGT nomenclature.
      allele_delim : delimiter separating the gene name from the allele number when asis_calls=True.
      build_asn : if True run tbl2asn on the generated .tbl and .fsa files.
      asn_template : template file (.sbt) to pass to tbl2asn.
      tbl2asn_exec : name of or path to the tbl2asn executable.
      format : input and output format.
      out_file : output file name without extension. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.

    Returns:
      tuple : the output (feature table, fasta) file names.
    """
    log = OrderedDict()
    log['START'] = 'ConvertDb'
    log['COMMAND'] = 'genbank'
    log['FILE'] = os.path.basename(db_file)
    printLog(log)

    # Define format operators
    try:
        reader, __, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)

    # Open input
    db_handle = open(db_file, 'rt')
    db_iter = reader(db_handle)

    # Check for required columns
    try:
        required = ['sequence_input',
                    'v_call', 'd_call', 'j_call',
                    'v_seq_start', 'd_seq_start', 'j_seq_start']
        checkFields(required, db_iter.fields, schema=schema)
    except LookupError as e:
        printError(e)

    # Open output
    if out_file is not None:
        out_name, __ = os.path.splitext(out_file)
        fsa_handle = open('%s.fsa' % out_name, 'w')
        tbl_handle = open('%s.tbl' % out_name, 'w')
    else:
        fsa_handle = getOutputHandle(db_file, out_label='genbank', out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'], out_type='fsa')
        tbl_handle = getOutputHandle(db_file, out_label='genbank', out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'], out_type='tbl')

    # Count records
    result_count = countDbFile(db_file)

    # Define writer
    writer = csv.writer(tbl_handle, delimiter='\t', quoting=csv.QUOTE_NONE)

    # Iterate over records
    start_time = time()
    rec_count, pass_count, fail_count = 0, 0, 0
    for rec in db_iter:
        # Print progress for previous iteration
        printProgress(rec_count, result_count, 0.05, start_time=start_time)
        rec_count += 1

        # Extract table dictionary
        name = None if asis_id else rec_count
        seq = makeGenbankSequence(rec, name=name, label=label, count_field=count_field, index_field=index_field,
                                  molecule=molecule, features=features)
        tbl = makeGenbankFeatures(rec, start=seq['start'], end=seq['end'], product=product,
                                  db_xref=db_xref, inference=inference, c_field=c_field,
                                  allow_stop=allow_stop, asis_calls=asis_calls, allele_delim=allele_delim)

        if tbl is not None:
            pass_count +=1
            # Write table
            writer.writerow(['>Features', seq['record'].id])
            for feature, qualifiers in tbl.items():
                writer.writerow(feature)
                if qualifiers:
                    for x in qualifiers:
                        writer.writerow(list(chain(['', '', ''], x)))

            # Write sequence
            SeqIO.write(seq['record'], fsa_handle, 'fasta')
        else:
            fail_count += 1

    # Final progress bar
    printProgress(rec_count, result_count, 0.05, start_time=start_time)

    # Run tbl2asn
    if build_asn:
        start_time = time()
        printMessage('Running tbl2asn', start_time=start_time, width=25)
        result = runASN(fsa_handle.name, template=asn_template, exec=tbl2asn_exec)
        printMessage('Done', start_time=start_time, end=True, width=25)

    # Print ending console log
    log = OrderedDict()
    log['OUTPUT_TBL'] = os.path.basename(tbl_handle.name)
    log['OUTPUT_FSA'] = os.path.basename(fsa_handle.name)
    log['RECORDS'] = rec_count
    log['PASS'] = pass_count
    log['FAIL'] = fail_count
    log['END'] = 'ConvertDb'
    printLog(log)

    # Close file handles
    tbl_handle.close()
    fsa_handle.close()
    db_handle.close()

    return (tbl_handle.name, fsa_handle.name)


def getArgParser():
    """
    Defines the ArgumentParser

    Arguments: 
    None
                      
    Returns: 
    an ArgumentParser object
    """
    # Define input and output field help message
    fields = dedent(
             '''
             output files:
                 airr
                     AIRR formatted database files.
                 changeo
                     Change-O formatted database files.
                 sequences
                     FASTA formatted sequences output from the subcommands fasta and clip.
                 genbank
                     feature tables and fasta files containing MiAIRR compliant input for tbl2asn.

             required fields:
                 sequence_id, sequence, sequence_alignment, junction, v_call, d_call, j_call, 
                 v_germline_start, v_germline_end, v_sequence_start, v_sequence_end, 
                 d_sequence_start, d_sequence_end, j_sequence_start, j_sequence_end 
                 
             optional fields:
                 germline_alignment, c_call, clone_id 
             ''')
    
    # Define ArgumentParser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            formatter_class=CommonHelpFormatter, add_help=False)
    group_help = parser.add_argument_group('help')
    group_help.add_argument('--version', action='version',
                            version='%(prog)s:' + ' %s %s' %(__version__, __date__))
    group_help.add_argument('-h', '--help', action='help', help='show this help message and exit')
    subparsers = parser.add_subparsers(title='subcommands', dest='command', metavar='',
                                       help='Database operation')
    # TODO:  This is a temporary fix for Python issue 9253
    subparsers.required = True

    # Define parent parsers
    default_parent = getCommonArgParser(failed=False, log=False, format=False)
    format_parent = getCommonArgParser(failed=False, log=False)

    # Subparser to convert changeo to AIRR files
    parser_airr = subparsers.add_parser('airr', parents=[default_parent],
                                        formatter_class=CommonHelpFormatter, add_help=False,
                                        help='Converts input to an AIRR TSV file.',
                                        description='Converts input to an AIRR TSV file.')
    parser_airr.set_defaults(func=convertToAIRR)

    # Subparser to convert AIRR to changeo files
    parser_changeo = subparsers.add_parser('changeo', parents=[default_parent],
                                       formatter_class=CommonHelpFormatter, add_help=False,
                                       help='Converts input into a Change-O TSV file.',
                                       description='Converts input into a Change-O TSV file.')
    parser_changeo.set_defaults(func=convertToChangeo)

    # Subparser to convert database entries to sequence file
    parser_fasta = subparsers.add_parser('fasta', parents=[default_parent],
                                       formatter_class=CommonHelpFormatter, add_help=False,
                                       help='Creates a fasta file from database records.',
                                       description='Creates a fasta file from database records.')
    group_fasta = parser_fasta.add_argument_group('conversion arguments')
    group_fasta.add_argument('--if', action='store', dest='id_field',
                              default=default_id_field,
                              help='The name of the field containing identifiers')
    group_fasta.add_argument('--sf', action='store', dest='seq_field',
                              default=default_seq_field,
                              help='The name of the field containing sequences')
    group_fasta.add_argument('--mf', nargs='+', action='store', dest='meta_fields',
                              help='List of annotation fields to add to the sequence description')
    parser_fasta.set_defaults(func=convertToFasta)
    
    # Subparser to convert database entries to clip-fasta file
    parser_baseln = subparsers.add_parser('baseline', parents=[default_parent],
                                          formatter_class=CommonHelpFormatter, add_help=False,
                                          description='Creates a BASELINe fasta file from database records.',
                                          help='''Creates a specially formatted fasta file
                                               from database records for input into the BASELINe
                                               website. The format groups clonally related sequences
                                               sequentially, with the germline sequence preceding
                                               each clone and denoted by headers starting with ">>".''')
    group_baseln = parser_baseln.add_argument_group('conversion arguments')
    group_baseln.add_argument('--if', action='store', dest='id_field',
                               default=default_id_field,
                               help='The name of the field containing identifiers')
    group_baseln.add_argument('--sf', action='store', dest='seq_field',
                               default=default_seq_field,
                               help='The name of the field containing reads')
    group_baseln.add_argument('--gf', action='store', dest='germ_field',
                               default=default_germ_field,
                               help='The name of the field containing germline sequences')
    group_baseln.add_argument('--cf', action='store', dest='cluster_field', default=None,
                               help='The name of the field containing containing sorted clone IDs')
    group_baseln.add_argument('--mf', nargs='+', action='store', dest='meta_fields',
                               help='List of annotation fields to add to the sequence description')
    parser_baseln.set_defaults(func=convertToBaseline)

    # Subparser to convert database entries to a GenBank fasta and feature table file
    parser_gb = subparsers.add_parser('genbank', parents=[format_parent],
                                       formatter_class=CommonHelpFormatter, add_help=False,
                                       help='Creates files for GenBank/TLS submissions.',
                                       description='Creates files for GenBank/TLS submissions.')
    # Genbank source information arguments
    group_gb_src = parser_gb.add_argument_group('source information arguments')
    group_gb_src.add_argument('--mol', action='store', dest='molecule', default=default_molecule,
                              help='''The source molecule type. Usually one of "mRNA" or "genomic DNA".''')
    group_gb_src.add_argument('--product', action='store', dest='product', default=default_product,
                              help='''The product name, such as "immunoglobulin heavy chain".''')
    group_gb_src.add_argument('--db', action='store', dest='db_xref', default=None,
                              help='''Name of the reference database used for alignment. 
                                   Usually "IMGT/GENE-DB".''')
    group_gb_src.add_argument('--inf', action='store', dest='inference', default=None,
                              help='''Name and version of the inference tool used for reference alignment in the 
                                   form tool:version.''')
    # Genbank sample information arguments
    group_gb_sam = parser_gb.add_argument_group('sample information arguments')
    group_gb_sam.add_argument('--organism', action='store', dest='organism', default=None,
                              help='The scientific name of the organism.')
    group_gb_sam.add_argument('--sex', action='store', dest='sex', default=None,
                              help='''If specified, adds the given sex annotation 
                                   to the fasta headers.''')
    group_gb_sam.add_argument('--isolate', action='store', dest='isolate', default=None,
                              help='''If specified, adds the given isolate annotation 
                                   (sample label) to the fasta headers.''')
    group_gb_sam.add_argument('--tissue', action='store', dest='tissue', default=None,
                              help='''If specified, adds the given tissue-type annotation 
                                   to the fasta headers.''')
    group_gb_sam.add_argument('--cell-type', action='store', dest='cell_type', default=None,
                              help='''If specified, adds the given cell-type annotation 
                                   to the fasta headers.''')
    group_gb_sam.add_argument('-y', action='store', dest='yaml_config', default=None,
                              help='''A yaml file specifying sample features (BioSample attributes) 
                                   in the form \'variable: value\'. If specified, any features provided in the 
                                   yaml file will override those provided at the commandline. Note,
                                   this config file applies to sample features only and
                                   cannot be used for required source features such as 
                                   the --product or --mol argument.''')
    # General genbank conversion arguments
    group_gb_cvt = parser_gb.add_argument_group('conversion arguments')
    group_gb_cvt.add_argument('--label', action='store', dest='label', default=None,
                              help='''If specified, add a field name to the sequence identifier. 
                                   Sequence identifiers will be output in the form <label>=<id>.''')
    group_gb_cvt.add_argument('--cf', action='store', dest='c_field', default=None,
                              help='''Field containing the C region call. If unspecified, the C region gene 
                                   call will be excluded from the feature table.''')
    group_gb_cvt.add_argument('--nf', action='store', dest='count_field', default=None,
                              help='''If specified, use the provided column to add the AIRR_READ_COUNT 
                                   note to the feature table.''')
    group_gb_cvt.add_argument('--if', action='store', dest='index_field', default=None,
                              help='''If specified, use the provided column to add the AIRR_CELL_INDEX 
                                   note to the feature table.''')
    group_gb_cvt.add_argument('--allow-stop', action='store_true', dest='allow_stop',
                              help='''If specified, retain records in the output with stop codons in the junction region.
                                   In such records the CDS will be removed and replaced with a similar misc_feature in 
                                   the feature table.''')
    group_gb_cvt.add_argument('--asis-id', action='store_true', dest='asis_id',
                              help='''If specified, use the existing sequence identifier for the output identifier. 
                                   By default, only the row number will be used as the identifier to avoid
                                   the 50 character limit.''')
    group_gb_cvt.add_argument('--asis-calls', action='store_true', dest='asis_calls',
                              help='''Specify to prevent alleles from being parsed using the IMGT nomenclature.
                                   Note, this requires the gene assignments to be exact matches to valid 
                                   records in the references database specified by the --db argument.''')
    group_gb_cvt.add_argument('--allele-delim', action='store', dest='allele_delim', default=default_allele_delim,
                              help='''The delimiter to use for splitting the gene name from the allele number.
                                   Note, this only applies when specifying --asis-calls. By default,
                                   this argument will be ignored and allele numbers extracted under the
                                   expectation of IMGT nomenclature consistency.''')
    group_gb_cvt.add_argument('--asn', action='store_true', dest='build_asn',
                              help='''If specified, run tbl2asn to generate the .sqn submission file after making 
                                   the .fsa and .tbl files.''')
    group_gb_cvt.add_argument('--sbt', action='store', dest='asn_template', default=None,
                              help='''If provided along with --asn, use the specified file for the template file
                                   argument to tbl2asn.''')
    group_gb_cvt.add_argument('--exec', action='store', dest='tbl2asn_exec', default=default_tbl2asn_exec,
                              help='The name or location of the tbl2asn executable.')
    parser_gb.set_defaults(func=convertToGenbank)

    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Genbank mode
    if args.command == 'genbank':
        # Create sample feature dictionary
        features = OrderedDict()
        if args_dict['organism'] is not None:
            features['organism'] = args_dict['organism']
        if args_dict['sex'] is not None:
            features['sex'] = args_dict['sex']
        if args_dict['isolate'] is not None:
            features['isolate'] = args_dict['isolate']
        if args_dict['tissue'] is not None:
            features['tissue_type'] = args_dict['tissue']
        if args_dict['cell_type'] is not None:
            features['cell_type'] = args_dict['cell_type']
        # Add yaml values
        if args_dict['yaml_config'] is not None:
            yaml_config = args_dict['yaml_config']
            if not os.path.exists(yaml_config):
                parser.error('%s does not exist' % yaml_config)
            else:
                features.update(yamlDict(yaml_config))
        # Update arguments
        args_dict['features'] = features
        # Clean arguments dictionary
        del args_dict['organism']
        del args_dict['sex']
        del args_dict['isolate']
        del args_dict['tissue']
        del args_dict['cell_type']
        del args_dict['yaml_config']

        # Check tbl2asn execution arguments
        if args_dict['build_asn']:
            template_file = args_dict['asn_template']
            tbl2asn_exec = args_dict['tbl2asn_exec']
            if not shutil.which(tbl2asn_exec):
                parser.error('%s does not exist or is not executable.' % tbl2asn_exec)
            if template_file is not None and not os.path.exists(template_file):
                parser.error('%s does not exist' % template_file)

    # Check argument pairs
    if args.command == 'add' and len(args_dict['fields']) != len(args_dict['values']):
        parser.error('You must specify exactly one value (-u) per field (-f)')
    elif args.command == 'rename' and len(args_dict['fields']) != len(args_dict['names']):
        parser.error('You must specify exactly one new name (-k) per field (-f)')
    elif args.command == 'update' and len(args_dict['values']) != len(args_dict['updates']):
        parser.error('You must specify exactly one value (-u) per replacement (-t)')

    # Clean arguments dictionary
    del args_dict['command']
    del args_dict['func']
    del args_dict['db_files']
    if 'out_files' in args_dict: del args_dict['out_files']

    # Call main function for each input file
    for i, f in enumerate(args.__dict__['db_files']):
        args_dict['db_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        args.func(**args_dict)



#!/usr/bin/env python3
"""
Assign Ig sequences into clones
"""

# Info
__author__ = 'Namita Gupta, Jason Anthony Vander Heiden, Gur Yaari, Mohamed Uduman'
from changeo import __version__, __date__

# Imports
import os
import re
import sys
from argparse import ArgumentParser
from collections import OrderedDict
from itertools import chain
from textwrap import dedent
from time import time
from Bio.Seq import translate

# Presto and changeo imports
from presto.Defaults import default_out_args
from presto.IO import printLog, printProgress, printCount, printWarning, printError
from presto.Multiprocessing import manageProcesses
from changeo.Defaults import default_format, default_v_field, default_j_field, default_junction_field, \
                             junction_attr, v_attr, j_attr
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs, \
                                setDefaultFields
from changeo.Distance import distance_models, calcDistances, formClusters
from changeo.IO import countDbFile, getDbFields, getFormatOperators, getOutputHandle, \
                       AIRRWriter, checkFields
from changeo.Multiprocessing import DbResult, feedDbQueue, processDbQueue

# Defaults
default_translate = False
default_distance = 0.0
default_index_mode = 'gene'
default_index_action = 'set'
default_distance_model = 'ham'
default_norm = 'len'
default_sym = 'avg'
default_linkage = 'single'
default_max_missing=0
choices_distance_model = ('ham', 'aa', 'hh_s1f', 'hh_s5f',
                          'mk_rs1nf', 'mk_rs5nf',
                          'hs1f_compat', 'm1n_compat')


def filterMissing(data, seq_field=junction_attr, v_field=v_attr,
                  j_field=j_attr, max_missing=default_max_missing):
    """
    Splits a set of sequences into passed and failed groups based on the number
    of missing characters in the sequence

    Arguments:
        data (changeo.Multiprocessing.DbData): data object.
        seq_field (str): Receptor sequence field to filter on.
        v_field (str): Receptor field containing the V call.
        j_field (str): Receptor field containing the J call.
        max_missing (int): maximum number of missing characters (non-ACGT) to permit before failing the record.

    Returns:
        changeo.Multiprocessing.DbResult : object containing filtered records.
    """
    # Function to validate the sequence string
    def _pass(seq):
        if len(seq) > 0 and len(re.findall(r'[^ACGT]', seq)) <= max_missing:
            return True
        else:
            return False

    # Define result object for iteration and get data records
    result = DbResult(data.id, data.data)

    if not data:
        result.data_pass = []
        result.data_fail = data.data
        return result

    result.data_pass = []
    result.data_fail = []
    for rec in data.data:
        seq = rec.getField(seq_field)
        if _pass(seq):  result.data_pass.append(rec)
        else:  result.data_fail.append(rec)

    # Add V(D)J to log
    result.log['ID'] = ','.join([str(x) for x in data.id])
    result.log['VCALL'] = ','.join(set([(r.getVAllele(field=v_field) or '') for r in data.data]))
    result.log['JCALL'] = ','.join(set([(r.getJAllele(field=j_field) or '') for r in data.data]))
    result.log['JUNCLEN'] = ','.join(set([(str(len(r.junction)) or '0') for r in data.data]))
    result.log['CLONED'] = len(result.data_pass)
    result.log['FILTERED'] = len(result.data_fail)

    return result


def indexByIdentity(index, key, rec, group_fields=None):
    """
    Updates a preclone index with a simple key

    Arguments:
      index : preclone index from groupByGene
      key : index key
      rec : Receptor to add to the index
      group_fields : additional annotation fields to use to group preclones;
                     if None use only V, J and junction length

    Returns:
      None : Updates index with new key and records.
    """
    index.setdefault(tuple(key), []).append(rec)


def indexByUnion(index, key, rec, group_fields=None):
    """
    Updates a preclone index with the union of nested keys

    Arguments:
      index : preclone index from groupByGene
      key : index key
      rec : Receptor to add to the index
      group_fields : additional annotation fields to use to group preclones;
                     if None use only V, J and junction length

    Returns:
      None : Updates index with new key and records.
    """
    # List of values for this/new key
    val = [rec]
    f_range = list(range(2, 3 + (len(group_fields) if group_fields else 0)))

    # See if field/junction length combination exists in index
    outer_dict = index
    for field in f_range:
        try:
            outer_dict = outer_dict[key[field]]
        except KeyError:
            outer_dict = None
            break
    # If field combination exists, look through Js
    j_matches = []
    if outer_dict is not None:
        for j in outer_dict.keys():
            if not set(key[1]).isdisjoint(set(j)):
                key[1] = tuple(set(key[1]).union(set(j)))
                j_matches += [j]
    # If J overlap exists, look through Vs for each J
    for j in j_matches:
        v_matches = []
        # Collect V matches for this J
        for v in outer_dict[j].keys():
            if not set(key[0]).isdisjoint(set(v)):
                key[0] = tuple(set(key[0]).union(set(v)))
                v_matches += [v]
        # If there are V overlaps for this J, pop them out
        if v_matches:
            val += list(chain(*(outer_dict[j].pop(v) for v in v_matches)))
            # If the J dict is now empty, remove it
            if not outer_dict[j]:
                outer_dict.pop(j, None)

    # Add value(s) into index nested dictionary
    # OMG Python pointers are the best!
    # Add field dictionaries into index
    outer_dict = index
    for field in f_range:
        outer_dict.setdefault(key[field], {})
        outer_dict = outer_dict[key[field]]
    # Add J, then V into index
    if key[1] in outer_dict:
        outer_dict[key[1]].update({key[0]: val})
    else:
        outer_dict[key[1]] = {key[0]: val}


def groupByGene(db_iter, group_fields=None, v_field=v_attr, j_field=j_attr,
                mode=default_index_mode, action=default_index_action):
    """
    Identifies preclonal groups by V, J and junction length

    Arguments: 
      db_iter : an iterator of Receptor objects defined by ChangeoReader
      group_fields : additional annotation fields to use to group preclones;
                     if None use only V, J and junction length
      v_field (str): Receptor attribute for V call
      j_field (str): Receptor attribute for J call
      mode : specificity of alignment call to use for assigning preclones;
             one of ('allele', 'gene')
      action : how to handle multiple value fields when assigning preclones;
               one of ('first', 'set')
    
    Returns: 
      dict: dictionary of {(V, J, junction length):[Receptor]}
    """
    # print(fields)
    # Define functions for grouping keys
    if mode == 'allele' and group_fields is None:
        def _get_key(rec, act):
            return [rec.getVAllele(act, field=v_field), rec.getJAllele(act, field=j_field),
                    None if rec.junction is None else len(rec.junction)]
    elif mode == 'gene' and group_fields is None:
        def _get_key(rec, act):  
            return [rec.getVGene(act, field=v_field), rec.getJGene(act, field=j_field),
                    None if rec.junction is None else len(rec.junction)]
    elif mode == 'allele' and group_fields is not None:
        def _get_key(rec, act):
            vdj = [rec.getVAllele(act, field=v_field), rec.getJAllele(act, field=j_field),
                    None if rec.junction is None else len(rec.junction)]
            ann = [rec.getField(k) for k in group_fields]
            return list(chain(vdj, ann))
    elif mode == 'gene' and group_fields is not None:
        def _get_key(rec, act):
            vdj = [rec.getVGene(act, field=v_field), rec.getJGene(act, field=j_field),
                    None if rec.junction is None else len(rec.junction)]
            ann = [rec.getField(k) for k in group_fields]
            return list(chain(vdj, ann))

    # Function to flatten nested dictionary
    def _flatten_dict(d, parent_key=''):
        items = []
        for k, v in d.items():
            new_key = parent_key + [k] if parent_key else [k]
            if isinstance(v, dict):
                items.extend(_flatten_dict(v, new_key).items())
            else:
                items.append((new_key, v))
        flat_dict = {None if None in i[0] else tuple(i[0]): i[1] for i in items}
        return flat_dict

    if action == 'first':
        index_func = indexByIdentity
    elif action == 'set':
        index_func = indexByUnion
    else:
        sys.stderr.write('Unrecognized action: %s.\n' % action)

    start_time = time()
    clone_index = {}
    rec_count = 0
    for rec in db_iter:
        key = _get_key(rec, action)

        # Print progress
        printCount(rec_count, step=1000, start_time=start_time, task='Grouping sequences')
        rec_count += 1

        # Assigned passed preclone records to key and failed to index None
        if all([k is not None and k != '' for k in key]):
            # Update index dictionary
            index_func(clone_index, key, rec, group_fields)
        else:
            clone_index.setdefault(None, []).append(rec)

    printCount(rec_count, step=1000, start_time=start_time, task='Grouping sequences', end=True)

    if action == 'set':
        clone_index = _flatten_dict(clone_index)

    return clone_index


def distanceClones(result, seq_field=default_junction_field, model=default_distance_model,
                   distance=default_distance, dist_mat=None, norm=default_norm, sym=default_sym,
                   linkage=default_linkage):
    """
    Separates a set of Receptor objects into clones

    Arguments:
      result : a changeo.Multiprocessing.DbResult object with filtered records to clone
      seq_field : sequence field used to calculate distance between records
      model : substitution model used to calculate distance
      distance : t  distance threshold to assign clonal groups
      dist_mat : pandas DataFrame of pairwise nucleotide or amino acid distances
      norm : normalization method
      sym : symmetry method
      linkage : type of linkage

    Returns:
      changeo.Multiprocessing.DbResult : an updated DbResult object
    """
    # Get distance matrix if not provided
    if dist_mat is None:
        try:
            dist_mat = distance_models[model]
        except KeyError:
            printError('Unrecognized distance model: %s' % args_dict['model'])

    # TODO:  can be cleaned up with abstract model class
    # Determine length of n-mers
    if model in ['hs1f_compat', 'm1n_compat', 'aa', 'ham', 'hh_s1f', 'mk_rs1nf']:
        nmer_len = 1
    elif model in ['hh_s5f', 'mk_rs5nf']:
        nmer_len = 5
    else:
        printError('Unrecognized distance model: %s.\n' % model)

    # Define unique junction mapping
    seq_map = {}
    for rec in result.data_pass:
        seq = rec.getField(seq_field)
        seq = re.sub(r'[\.-]', 'N', seq)
        # Translate sequence for amino acid model
        if model == 'aa':
            # Check for valid translation
            if len(seq) % 3 > 0:  seq = seq + 'N' * (3 - len(seq) % 3)
            seq = translate(seq)
        seq_map.setdefault(seq, []).append(rec)

    # Define sequences
    sequences = list(seq_map.keys())

    # Zero record case
    if not sequences:
        result.valid = False
        result.log['CLONES'] = 0
        return result

    # Single record case
    if len(sequences) == 1:
        result.results = {1: result.data_pass}
        result.valid = True
        result.log['CLONES'] = 1
        return result

    # Calculate pairwise distance matrix
    dists = calcDistances(sequences, nmer_len, dist_mat, sym=sym, norm=norm)

    # Perform hierarchical clustering
    clusters = formClusters(dists, linkage, distance)

    # Turn clusters into clone dictionary
    clone_dict = {}
    for i, c in enumerate(clusters):
        clone_dict.setdefault(c, []).extend(seq_map[sequences[i]])

    if clone_dict:
        result.results = clone_dict
        result.valid = True
        result.log['CLONES'] = len(clone_dict)
    else:
        result.log['CLONES'] = 0

    return result


def collectQueue(alive, result_queue, collect_queue, db_file, fields,
                 writer=AIRRWriter, out_file=None, out_args=default_out_args):
    """
    Assembles results from a queue of individual sequence results and manages log/file I/O

    Arguments:
      alive = a multiprocessing.Value boolean controlling whether processing continues
              if False exit process
      result_queue : a multiprocessing.Queue holding processQueue results
      collect_queue : a multiprocessing.Queue to store collector return values
      db_file : the input database file name
      fields : list of output field names
      writer : writer class.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs
    
    Returns:
       None : Adds a dictionary with key value pairs to collect_queue containing
            'log' defining a log object along with the 'pass' and 'fail' output file names.
    """
    # Wrapper for opening handles and writers
    def _open(x, f, writer=writer, out_file=out_file):
        if out_file is not None and x == 'pass':
            handle = open(out_file, 'w')
        else:
            handle = getOutputHandle(db_file,
                                     out_label='clone-%s' % x,
                                     out_dir=out_args['out_dir'],
                                     out_name=out_args['out_name'],
                                     out_type=out_args['out_type'])
        return handle, writer(handle, fields=f)

    # Open log file
    try:
        # Count input records
        result_count = countDbFile(db_file)

        # Define log handle
        if out_args['log_file'] is None:  
            log_handle = None
        else:  
            log_handle = open(out_args['log_file'], 'w')
    except:
        #sys.stderr.write('Exception in collector file opening step\n')
        alive.value = False
        raise

    # Get results from queue and write to files
    try:
        # Initialize handles, writers and counters
        pass_handle, pass_writer = None, None
        fail_handle, fail_writer = None, None
        rec_count, clone_count, pass_count, fail_count = 0, 0, 0, 0
        start_time = time()

        # Iterator over results queue until sentinel object reached
        while alive.value:
            # Get result from queue
            if result_queue.empty():  continue
            else:  result = result_queue.get()
            # Exit upon reaching sentinel
            if result is None:  break

            # Print progress for previous iteration and update record count
            printProgress(rec_count, result_count, 0.05, start_time=start_time, task='Assigning clones')
            rec_count += len(result.data)
            
            # Write passed and failed records
            if result:
                # Writing passing sequences
                for clone in result.results.values():
                    clone_count += 1
                    for i, rec in enumerate(clone, start=1):
                        pass_count += 1
                        rec.setField('clone', str(clone_count))
                        result.log['CLONE%i-%i' % (clone_count, i)] = rec.junction
                        try:
                            pass_writer.writeReceptor(rec)
                        except AttributeError:
                            # Open pass file and define writer object
                            pass_handle, pass_writer = _open('pass', fields)
                            pass_writer.writeReceptor(rec)

                # Write failed sequences from passing sets
                if result.data_fail:
                    # Write failed sequences
                    for i, rec in enumerate(result.data_fail, start=1):
                        fail_count += 1
                        result.log['FAIL%i-%i' % (clone_count, i)] = rec.junction
                        if out_args['failed']:
                            try:
                                fail_writer.writeReceptor(rec)
                            except AttributeError:
                                # Open fail file and define writer object
                                fail_handle, fail_writer = _open('fail', fields)
                                fail_writer.writeReceptor(rec)
            else:
                # Write failing records
                for i, rec in enumerate(result.data, start=1):
                    fail_count += 1
                    result.log['CLONE0-%i' % (i)] = rec.junction
                    if out_args['failed']:
                        try:
                            fail_writer.writeReceptor(rec)
                        except AttributeError:
                            # Open fail file and define writer object
                            fail_handle, fail_writer = _open('fail', fields)
                            fail_writer.writeReceptor(rec)
                    
            # Write log
            printLog(result.log, handle=log_handle)
        else:
            sys.stderr.write('PID %s>  Error in sibling process detected. Cleaning up.\n' \
                             % os.getpid())
            return None
        
        # Print total counts
        printProgress(rec_count, result_count, 0.05, start_time=start_time, task='Assigning clones')

        # Update return list
        log = OrderedDict()
        log['OUTPUT'] = os.path.basename(pass_handle.name) if pass_handle is not None else None
        log['CLONES'] = clone_count
        log['RECORDS'] = rec_count
        log['PASS'] = pass_count
        log['FAIL'] = fail_count

        # Close file handles and generate return data
        collect_dict = {'log': log, 'pass': None, 'fail': None}
        if pass_handle is not None:
            collect_dict['pass'] = pass_handle.name
            pass_handle.close()
        if fail_handle is not None:
            collect_dict['fail'] = fail_handle.name
            fail_handle.close()
        if log_handle is not None:
            log_handle.close()
        collect_queue.put(collect_dict)
    except:
        alive.value = False
        raise

    return None


def defineClones(db_file, seq_field=default_junction_field, v_field=default_v_field,
                 j_field=default_j_field, max_missing=default_max_missing,
                 group_fields=None, group_func=groupByGene, group_args={},
                 clone_func=distanceClones, clone_args={},
                 format=default_format, out_file=None, out_args=default_out_args,
                 nproc=None, queue_size=None):
    """
    Define clonally related sequences
    
    Arguments:
      db_file : filename of input database.
      seq_field : sequence field used to determine clones.
      v_field : field containing the V call.
      j_field : field containing the J call.
      max_missing : maximum number of non-ACGT characters to allow in the junction sequence.
      group_fields : additional annotation fields to use to group preclones;
                     if None use only V and J.
      group_func : the function to use for assigning preclones.
      group_args : a dictionary of arguments to pass to group_func.
      clone_func : the function to use for determining clones within preclonal groups.
      clone_args : a dictionary of arguments to pass to clone_func.
      format : input and output format.
      out_file : output file name. Automatically generated from the input file if None.
      out_args : common output argument dictionary from parseCommonArgs.
      nproc : the number of processQueue processes;
              if None defaults to the number of CPUs.
      queue_size : maximum size of the argument queue;
                   if None defaults to 2*nproc.
    
    Returns:
      dict: dictionary of output pass and fail files.
    """
    # Print parameter info
    log = OrderedDict()
    log['START'] = 'DefineClones'
    log['FILE'] = os.path.basename(db_file)
    log['SEQ_FIELD'] = seq_field
    log['V_FIELD'] = v_field
    log['J_FIELD'] = j_field
    log['MAX_MISSING'] = max_missing
    log['GROUP_FIELDS'] = ','.join(group_fields) if group_fields is not None else None
    for k in sorted(group_args):
        log[k.upper()] = group_args[k]
    for k in sorted(clone_args):
        if k != 'dist_mat':  log[k.upper()] = clone_args[k]
    log['NPROC'] = nproc
    printLog(log)

    # Define format operators
    try:
        reader, writer, schema = getFormatOperators(format)
    except ValueError:
        printError('Invalid format %s.' % format)

    # Translate to Receptor attribute names
    seq_field = schema.toReceptor(seq_field)
    v_field = schema.toReceptor(v_field)
    j_field = schema.toReceptor(j_field)
    if group_fields is not None:
        group_fields = [schema.toReceptor(f) for f in group_fields]

    # Define feeder function and arguments
    group_args['group_fields'] = group_fields
    group_args['v_field'] = v_field
    group_args['j_field'] = j_field
    feed_args = {'db_file': db_file,
                 'reader': reader,
                 'group_func': group_func,
                 'group_args': group_args}

    # Define worker function and arguments
    filter_args = {'seq_field': seq_field,
                   'v_field': v_field,
                   'j_field': j_field,
                   'max_missing': max_missing}
    clone_args['seq_field'] = seq_field
    work_args = {'process_func': clone_func,
                 'process_args': clone_args,
                 'filter_func': filterMissing,
                 'filter_args': filter_args}

    # Define collector function and arguments
    out_fields = getDbFields(db_file, add=schema.fromReceptor('clone'), reader=reader)
    out_args['out_type'] = schema.out_type
    collect_args = {'db_file': db_file,
                    'fields': out_fields,
                    'writer': writer,
                    'out_file': out_file,
                    'out_args': out_args}

    # Check for required columns
    try:
        required = ['junction']
        checkFields(required, out_fields, schema=schema)
    except LookupError as e:
        printError(e)

    # Call process manager
    result = manageProcesses(feed_func=feedDbQueue, work_func=processDbQueue, collect_func=collectQueue,
                             feed_args=feed_args, work_args=work_args, collect_args=collect_args,
                             nproc=nproc, queue_size=queue_size)
        
    # Print log
    result['log']['END'] = 'DefineClones'
    printLog(result['log'])
    output = {k: v for k, v in result.items() if k in ('pass', 'fail')}

    return output


def getArgParser():
    """
    Defines the ArgumentParser

    Arguments: 
    None
                      
    Returns: 
    an ArgumentParser object
    """
    # Define input and output fields
    fields = dedent(
             '''
             output files:
                 clone-pass
                     database with assigned clonal group numbers.
                 clone-fail
                     database with records failing clonal grouping.

             required fields:
                 sequence_id, v_call, j_call, junction
                
             output fields:
                 clone_id
             ''')
    # Define argument parser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            parents=[getCommonArgParser(format=True, multiproc=True)],
                            formatter_class=CommonHelpFormatter, add_help=False)

    # Distance cloning method
    group = parser.add_argument_group('cloning arguments')
    group.add_argument('--sf', action='store', dest='seq_field', default=None,
                        help='''Field to be used to calculate distance between records.
                              Defaults to junction (airr) or JUNCTION (changeo).''')
    group.add_argument('--vf', action='store', dest='v_field', default=None,
                        help='''Field containing the germline V segment call.
                             Defaults to v_call (airr) or V_CALL (changeo).''')
    group.add_argument('--jf', action='store', dest='j_field', default=None,
                        help='''Field containing the germline J segment call.
                             Defaults to j_call (airr) or J_CALL (changeo).''')
    group.add_argument('--gf', nargs='+', action='store', dest='group_fields', default=None,
                        help='Additional fields to use for grouping clones aside from V, J and junction length.')
    group.add_argument('--mode', action='store', dest='mode',
                        choices=('allele', 'gene'), default=default_index_mode,
                        help='''Specifies whether to use the V(D)J allele or gene for
                             initial grouping.''')
    group.add_argument('--act', action='store', dest='action',
                        choices=('first', 'set'), default=default_index_action,
                        help='''Specifies how to handle multiple V(D)J assignments for initial grouping. 
                             The "first" action will use only the first gene listed.
                             The "set" action will use all gene assignments and construct a larger gene
                             grouping composed of any sequences sharing an assignment or linked to another
                             sequence by a common assignment (similar to single-linkage).''')
    group.add_argument('--model', action='store', dest='model',
                        choices=choices_distance_model,
                        default=default_distance_model,
                        help='''Specifies which substitution model to use for calculating distance
                             between sequences. The "ham" model is nucleotide Hamming distance and
                             "aa" is amino acid Hamming distance. The "hh_s1f" and "hh_s5f" models are
                             human specific single nucleotide and 5-mer content models, respectively,
                             from Yaari et al, 2013. The "mk_rs1nf" and "mk_rs5nf" models are
                             mouse specific single nucleotide and 5-mer content models, respectively,
                             from Cui et al, 2016. The "m1n_compat" and "hs1f_compat" models are
                             deprecated models provided backwards compatibility with the "m1n" and
                             "hs1f" models in Change-O v0.3.3 and SHazaM v0.1.4. Both
                             5-mer models should be considered experimental.''')
    group.add_argument('--dist', action='store', dest='distance', type=float,
                        default=default_distance,
                        help='The distance threshold for clonal grouping')
    group.add_argument('--norm', action='store', dest='norm',
                        choices=('len', 'mut', 'none'), default=default_norm,
                        help='''Specifies how to normalize distances. One of none
                             (do not normalize), len (normalize by length),
                             or mut (normalize by number of mutations between sequences).''')
    group.add_argument('--sym', action='store', dest='sym',
                        choices=('avg', 'min'), default=default_sym,
                        help='''Specifies how to combine asymmetric distances. One of avg
                             (average of A->B and B->A) or min (minimum of A->B and B->A).''')
    group.add_argument('--link', action='store', dest='linkage',
                        choices=('single', 'average', 'complete'), default=default_linkage,
                        help='''Type of linkage to use for hierarchical clustering.''')
    group.add_argument('--maxmiss', action='store', dest='max_missing', type=int,
                        default=default_max_missing,
                        help='''The maximum number of non-ACGT characters (gaps or Ns) to 
                             permit in the junction sequence before excluding the record 
                             from clonal assignment. Note, under single linkage 
                             non-informative positions can create artifactual links 
                             between unrelated sequences. Use with caution.''')
    parser.set_defaults(group_func=groupByGene)
    parser.set_defaults(clone_func=distanceClones)
        
    return parser


if __name__ == '__main__':
    """
    Parses command line arguments and calls main function
    """
    # Parse arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)

    # Set default fields
    default_fields = {'seq_field': default_junction_field,
                      'v_field': default_v_field,
                      'j_field': default_j_field}
    args_dict = setDefaultFields(args_dict, default_fields, format=args_dict['format'])

    # Define grouping and cloning function arguments
    args_dict['group_args'] = {'action': args_dict['action'],
                               'mode':args_dict['mode']}
    args_dict['clone_args'] = {'model':  args_dict['model'],
                               'distance':  args_dict['distance'],
                               'norm': args_dict['norm'],
                               'sym': args_dict['sym'],
                               'linkage': args_dict['linkage']}

    # Get distance matrix
    try:
        args_dict['clone_args']['dist_mat'] = distance_models[args_dict['model']]
    except KeyError:
        printError('Unrecognized distance model: %s' % args_dict['model'])

    # Clean argument dictionary
    del args_dict['action']
    del args_dict['mode']
    del args_dict['model']
    del args_dict['distance']
    del args_dict['norm']
    del args_dict['sym']
    del args_dict['linkage']

    # Clean arguments dictionary
    del args_dict['db_files']
    if 'out_files' in args_dict: del args_dict['out_files']

    # Call main function for each input file
    for i, f in enumerate(args.__dict__['db_files']):
        args_dict['db_file'] = f
        args_dict['out_file'] = args.__dict__['out_files'][i] \
            if args.__dict__['out_files'] else None
        defineClones(**args_dict)



#!/usr/bin/env python3
"""
Converts TSV files into IgPhyML input files
"""

# Info
__author__ = "Kenneth Hoehn"
from changeo import __version__, __date__

# Imports
import os
import random
import subprocess
import multiprocessing as mp
from argparse import ArgumentParser
from collections import OrderedDict
from textwrap import dedent
from time import time
from Bio.Seq import Seq
from functools import partial

# Presto and changeo imports
from presto.Defaults import default_out_args
from presto.IO import  printLog, printMessage, printWarning, printError, printDebug
from changeo.Defaults import default_format
from changeo.IO import splitName, getDbFields, getFormatOperators, getOutputHandle, getOutputName
from changeo.Alignment import RegionDefinition
from changeo.Commandline import CommonHelpFormatter, checkArgs, getCommonArgParser, parseCommonArgs


def correctMidCodonStart(scodons, qi, debug):
    """
    Find and mask split codons

    Arguments:
      scodons (list): list of codons in IMGT sequence.
      qi (str) : input sequence.
      spos (int) : starting position of IMGT sequence in input sequence.
      debug (bool) : print debugging statements.

    Returns:
      tuple: (modified input sequence, modified starting position of IMGT sequence in input sequence).
    """
    spos = 0   
    for i in range(0, len(scodons)):
        printDebug("%s %s" % (scodons[i], qi[0:3]), debug)
        if scodons[i] != "...":
            if scodons[i][0:2] == "..":
                scodons[i] = "NN" + scodons[i][2]
                #sometimes IMGT will just cut off first letter if non-match, at which point we"ll just want to mask the
                #first codon in the IMGT seq, other times it will be legitimately absent from the query, at which point
                #we have to shift the frame. This attempts to correct for this by looking at the next codon over in the
                #alignment
                if scodons[i][2:3] != qi[2:3] or scodons[i + 1] != qi[3:6]:
                    qi = "NN" + qi
                spos = i
                break
            elif scodons[i][0] == ".":
                scodons[i] = "N" + scodons[i][1:3]
                if scodons[i][1:3] != qi[1:3] or scodons[i+1] != qi[3:6]:
                    qi = "N" + qi
                spos = i
                break
            else:
                spos = i
                break

    return qi, spos


def checkFrameShifts(receptor, oqpos, ospos, log, debug):
    """
    Checks whether a frameshift occured in a sequence

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object.
      oqpos (int) : position of interest in input sequence.
      ospos (int) : position of interest in IMGT sequence.
      log (dict) : log of information for each sequence.
      debug (bool) : print debugging statements.
    """   
    frameshifts = 0
    for ins in range(1, 3):
        ros = receptor.sequence_input
        ris = receptor.sequence_imgt
        psite = receptor.v_seq_start - 1 + oqpos*3
        pisite = ospos * 3
        if (psite + 3 + ins) < len(ros) and (pisite + 3) < len(ris):
        #cut out 1 or 2 nucleotides downstream of offending codon
            receptor.sequence_input = ros[0:(psite + 3)] + ros[(psite + 3 + ins):]
            receptor.sequence_imgt = ris[0:(pisite + 3)] + ris[(pisite + 3):]

            # Debug sequence modifications
            printDebug(ros, debug)
            printDebug(receptor.sequence_input, debug)
            printDebug(ris, debug)
            printDebug(receptor.sequence_imgt, debug)
            printDebug("RUNNING %d" % ins, debug)

            mout = maskSplitCodons(receptor, recursive=True)
            if mout[1]["PASS"]:
                #if debug:
                receptor.sequence_input = ros
                receptor.sequence_imgt = ris
                frameshifts += 1
                printDebug("FRAMESHIFT of length %d!" % ins, debug)
                log["FAIL"] = "SINGLE FRAME-SHIFTING INSERTION"
                break
            else:
                receptor.sequence_input = ros
                receptor.sequence_imgt = ris

    return frameshifts


def findAndMask(receptor, scodons, qcodons, spos, s_end, qpos, log, debug, recursive=False):
    """
    Find and mask split codons

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object.
      scodons (list): list of codons in IMGT sequence
      qcodons (list): list of codons in input sequence
      spos (int): starting position of IMGT sequence in input sequence
      s_end (int): end of IMGT sequence
      qpos (int): starting position of input sequence in IMGT sequence
      log (dict): log of information for each sequence
      debug (bool): print debugging statements?
      recursive (bool): was this function called recursively?
    """
    frameshifts = 0
    while spos < s_end and qpos < len(qcodons):
        if debug:
            print(scodons[spos] + "\t" + qcodons[qpos])
        if scodons[spos] == "..." and qcodons[qpos] != "...": #if IMGT gap, move forward in imgt
            spos += 1
        elif scodons[spos] == qcodons[qpos]: # if both are the same, move both forward
            spos += 1
            qpos += 1
        elif qcodons[qpos] == "N": # possible that SEQ-IMGT ends on a bunch of Ns
            qpos += 1
            spos += 1
        else: # if not the same, mask IMGT at that site and scan forward until you find a codon that matches next site
            if debug:
                print("checking %s at position %d %d" % (scodons[spos], spos, qpos))
            ospos=spos
            oqpos=qpos
            spos += 1
            qpos += 1
            while spos < s_end and scodons[spos] == "...": #possible next codon is just a gap
                spos += 1
            while qpos < len(qcodons) and spos < s_end and scodons[spos] != qcodons[qpos]:
                printDebug("Checking " + scodons[spos]+ "\t" + qcodons[qpos], debug)
                qpos += 1
            if qcodons[qpos-1] == scodons[ospos]: #if codon in previous position is equal to original codon, it was preserved
                qpos -= 1
                spos = ospos
                printDebug("But codon was apparently preserved", debug)
                if "IN-FRAME" in log:
                    log["IN-FRAME"] = log["IN-FRAME"] + "," +  str(spos)
                else:
                    log["IN-FRAME"] = str(spos)
            elif qpos >= len(qcodons) and spos < s_end:
                printDebug("FAILING MATCH", debug)
                log["PASS"] = False #if no match for the adjacent codon was found, something"s up.
                log["FAIL"] = "FAILED_MATCH_QSTRING:"+str(spos)
                #figure out if this was due to a frame-shift by repeating this method but with an edited input sequence
                if not recursive:
                    frameshifts += checkFrameShifts(receptor, oqpos, ospos, log, debug)

            elif spos >= s_end or qcodons[qpos] != scodons[spos]:
                scodons[ospos] = "NNN"
                if spos >= s_end:
                    printDebug("Masked %s at position %d, at end of subject sequence" % (scodons[ospos], ospos), debug)
                    if "END-MASKED" in log:
                        log["END-MASKED"] = log["END-MASKED"] + "," + str(spos)
                    else:
                        log["END-MASKED"] = str(spos)
                else:
                    printDebug("Masked %s at position %d, but couldn't find upstream match" % (scodons[ospos], ospos), debug)
                    log["PASS"]=False
                    log["FAIL"]="FAILED_MATCH:"+str(spos)
            elif qcodons[qpos] == scodons[spos]:
                printDebug("Masked %s at position %d" % (scodons[ospos], ospos), debug)
                scodons[ospos] = "NNN"
                if "MASKED" in  log:
                    log["MASKED"] = log["MASKED"] + "," + str(spos)
                else:
                    log["MASKED"] = str(spos)
            else:
                log["PASS"] = False
                log["FAIL"] = "UNKNOWN"


def maskSplitCodons(receptor, recursive=False, mask=True):
    """
    Identify junction region by IMGT definition.

    Arguments:
      receptor (changeo.Receptor.Receptor): Receptor object.
      recursive (bool) : was this method part of a recursive call?
      mask (bool) : mask split codons for use with igphyml?

    Returns:
      str: modified IMGT gapped sequence.
      log: dict of sequence information
    """
    debug = False
    qi = receptor.sequence_input
    si = receptor.sequence_imgt
    log = OrderedDict()
    log["ID"]=receptor.sequence_id
    log["CLONE"]=receptor.clone
    log["PASS"] = True
    if debug:
        print(receptor.sequence_id)
    # adjust starting position of query sequence
    qi = qi[(receptor.v_seq_start - 1):]

    #tally where --- gaps are in IMGT sequence and remove them for now
    gaps = []
    ndotgaps = []
    nsi = ""
    for i in range(0,len(si)):
        if si[i] == "-":
            gaps.append(1)
            ndotgaps.append(1)
        else:
            gaps.append(0)
            nsi = nsi + si[i]
            if si[i] != ".":
                ndotgaps.append(0)

    #find any gaps not divisible by three
    curgap = 0
    for i in ndotgaps:
        if i == 1:
            curgap += 1
        elif i == 0 and curgap != 0:
            if curgap % 3 != 0 :
                printDebug("Frame-shifting gap detected! Refusing to include sequence.", debug)
                log["PASS"] = False
                log["FAIL"] = "FRAME-SHIFTING DELETION"
                log["SEQ_IN"] = receptor.sequence_input
                log["SEQ_IMGT"] = receptor.sequence_imgt
                log["SEQ_MASKED"] = receptor.sequence_imgt
                return receptor.sequence_imgt, log
            else:
                curgap = 0
    si = nsi

    scodons = [si[i:i + 3] for i in range(0, len(si), 3)]

    # deal with the fact that it's possible to start mid-codon
    qi,spos = correctMidCodonStart(scodons, qi, debug)

    qcodons = [qi[i:i + 3] for i in range(0, len(qi), 3)]

    frameshifts = 0
    s_end = 0 #adjust for the fact that IMGT sequences can end on gaps
    for i in range(spos, len(scodons)):
        if scodons[i] != "..." and len(scodons[i]) == 3 and scodons[i] != "NNN":
            s_end = i

    printDebug("%i:%i:%s" % (s_end, len(scodons), scodons[s_end]), debug)
    s_end += 1
    qpos = 0

    if mask:
        findAndMask(receptor, scodons, qcodons, spos, s_end, qpos, log, debug, recursive)

    if not log["PASS"] and not recursive:
        log["FRAMESHIFTS"] = frameshifts
    if len(scodons[-1]) != 3:
        if scodons[-1] == ".." or scodons[-1] == ".":
            scodons[-1] = "..."
        else:
            scodons[-1] = "NNN"
        if "END-MASKED" in log:
            log["END-MASKED"] = log["END-MASKED"] + "," + str(len(scodons))
        else:
            log["END-MASKED"] = str(spos)

    concatenated_seq = Seq("")
    for i in scodons:
        concatenated_seq += i

    # add --- gaps back to IMGT sequence
    ncon_seq = ""
    counter = 0
    for i in gaps:
        #print(str(i) + ":" + ncon_seq)
        if i == 1:
            ncon_seq = ncon_seq + "."
        elif i == 0:
            ncon_seq = ncon_seq + concatenated_seq[counter]
            counter += 1
    ncon_seq = ncon_seq + concatenated_seq[counter:]
    concatenated_seq = ncon_seq
    log["SEQ_IN"] = receptor.sequence_input
    log["SEQ_IMGT"] = receptor.sequence_imgt
    log["SEQ_MASKED"] = concatenated_seq

    return concatenated_seq, log


def unAmbigDist(seq1, seq2, fbreak=False):
    """
    Calculate the distance between two sequences counting only A,T,C,Gs

    Arguments:
      seq1 (str): sequence 1
      seq2 (str): sequence 2
      fbreak (bool): break after first difference found?

    Returns:
      int: number of ACGT differences.
    """
    if len(seq1) != len(seq2):
        printError("Sequences are not the same length! %s %s" % (seq1, seq2))

    dist = 0
    for i in range(0,len(seq1)):
        if seq1[i] != "N" and seq1[i] != "-" and seq1[i] != ".":
            if seq2[i] != "N" and seq2[i] != "-" and seq2[i] != ".":
                if seq1[i] != seq2[i]:
                    dist += 1
                    if fbreak:
                        break

    return dist


def deduplicate(useqs, receptors, log=None, meta_data=None, delim=":"):
    """
    Collapses identical sequences

    Argument:
      useqs (dict): unique sequences within a clone. maps sequence to index in Receptor list.
      receptors (dict): receptors within a clone (index is value in useqs dict).
      log (collections.OrderedDict): log of sequence errors.
      meta_data (str): Field to append to sequence IDs. Splits identical sequences with different meta_data.
      meta_data (str): Field to append to sequence IDs. Splits identical sequences with different meta_data.
      delim (str): delimited to use when appending meta_data.

    Returns:
      list: deduplicated receptors within a clone.
    """

    keys = list(useqs.keys())
    join = {} # id -> sequence id to join with (least ambiguous chars)
    joinseqs = {} # id -> useq to join with (least ambiguous chars)
    ambigchar = {} #sequence id -> number ATCG nucleotides
    for i in range(0,len(keys)-1):
        for j in range(i+1,len(keys)):
            ki = keys[i]
            kj = keys[j]
            if meta_data is None:
                ski = keys[i]
                skj = keys[j]
            else:
                ski, cid = keys[i].split(delim)
                skj, cid = keys[j].split(delim)
            ri = receptors[useqs[ki]]
            rj = receptors[useqs[kj]]
            dist = unAmbigDist(ski, skj, True)
            m_match = True
            if meta_data is not None:
                matches = 0
                for m in meta_data:
                    if ri.getField(m) == rj.getField(m) and m != "DUPCOUNT":
                        matches += 1
                m_match = (matches == len(meta_data))
            if dist == 0 and m_match:
                ncounti = ki.count("A") + ki.count("T") + ki.count("G") + ki.count("C")
                ncountj = kj.count("A") + kj.count("T") + kj.count("G") + kj.count("C")
                ambigchar[useqs[ki]] = ncounti
                ambigchar[useqs[kj]] = ncountj
                # this algorithm depends on the fact that all sequences are compared pairwise, and all are zero
                # distance from the sequence they will be collapse to.
                if ncountj > ncounti:
                    nci = 0
                    if useqs[ki] in join:
                        nci = ambigchar[join[useqs[ki]]]
                    if nci < ncountj:
                        join[useqs[ki]] = useqs[kj]
                        joinseqs[ki] = kj
                else:
                    ncj = 0
                    if useqs[kj] in join:
                        ncj = ambigchar[join[useqs[kj]]]
                    if ncj < ncounti:
                        join[useqs[kj]] = useqs[ki]
                        joinseqs[kj] = ki

    # loop through list of joined sequences and collapse
    keys = list(useqs.keys())
    for k in keys:
        if useqs[k] in join:
            rfrom = receptors[useqs[k]]
            rto = receptors[join[useqs[k]]]
            rto.dupcount += rfrom.dupcount
            if log is not None:
                log[rfrom.sequence_id]["PASS"] = False
                log[rfrom.sequence_id]["DUPLICATE"] = True
                log[rfrom.sequence_id]["COLLAPSETO"] = joinseqs[k]
                log[rfrom.sequence_id]["COLLAPSEFROM"] = k
                log[rfrom.sequence_id]["FAIL"] = "Collapsed with " + rto.sequence_id
            del useqs[k]

    return useqs


def hasPTC(sequence):
    """
    Determines whether a PTC exits in a sequence

    Arguments:
      sequence (str): IMGT gapped sequence in frame 1.

    Returns:
      int: negative if not PTCs, position of PTC if found.

    """
    ptcs = ("TAA", "TGA", "TAG", "TRA", "TRG", "TAR", "TGR", "TRR")
    for i in range(0, len(sequence), 3):
            if sequence[i:(i+3)] in ptcs:
                return i
    return -1



def rmCDR3(sequences, clones):
    """
    Remove CDR3 from all sequences and germline of a clone

     Arguments:
       sequences (list): list of sequences in clones.
       clones (list): list of Receptor objects.
    """

    for i in range(0,len(sequences)):
        imgtar = clones[i].getField("imgtpartlabels")
        germline = clones[i].getField("germline_imgt_d_mask")
        nseq = []
        nimgtar = []
        ngermline = []
        ncdr3 = 0
        #print("imgtarlen: " + str(len(imgtar)))
        #print("seqlen: " + str(len(sequences[i])))
        #print("germline: " + str(len(germline)))
        #if len(germline) < len(sequences[i]):
        #    print("\n" + str(clones[i].sequence_id))
        #    print("\n " + str((sequences[i])) )
        #    print("\n" + str((germline)))
        for j in range(0,len(imgtar)):
            if imgtar[j] != 108:
                nseq.append(sequences[i][j])
                if j < len(germline):
                    ngermline.append(germline[j])
                nimgtar.append(imgtar[j])
            else:
                ncdr3 += 1
        clones[i].setField("imgtpartlabels",nimgtar)
        clones[i].setField("germline_imgt_d_mask", "".join(ngermline))
        sequences[i] = "".join(nseq)
        #print("Length: " + str(ncdr3))


def characterizePartitionErrors(sequences, clones, meta_data):
    """
    Characterize potential mismatches between IMGT labels within a clone

    Arguments:
      sequences (list): list of sequences in clones.
      clones (list): list of Receptor objects.
      meta_data (str): Field to append to sequence IDs. Splits identical sequences with different meta_data.
        
    Returns:
      tuple: tuple of length four containing a list of IMGT positions for first sequence in clones,
             the germline sequence of the first receptor in clones, the length of the first sequence in clones,
             and the number of sequences in clones.
    """
    sites = len(sequences[0])
    nseqs = len(sequences)
    imgtar = clones[0].getField("imgtpartlabels")
    germline = clones[0].getField("germline_imgt_d_mask")
    if germline == "":
        germline = clones[0].getField("germline_imgt")

    correctseqs = False

    for seqi in range(0, len(sequences)):
        i = sequences[seqi]
        if len(i) != sites or len(clones[seqi].getField("imgtpartlabels")) != len(imgtar):
            correctseqs = True

    if correctseqs:
        maxlen = sites
        maximgt = len(imgtar)
        for j in range(0,len(sequences)):
            if len(sequences[j]) > maxlen:
                maxlen = len(sequences[j])
            if len(clones[j].getField("imgtpartlabels")) > maximgt:
                imgtar = clones[j].getField("imgtpartlabels")
                maximgt = len(imgtar)
        sites = maxlen
        for j in range(0,len(sequences)):
            cimgt = clones[j].getField("imgtpartlabels")
            seqdiff = maxlen - len(sequences[j])
            imgtdiff = len(imgtar)-len(cimgt)
            sequences[j] = sequences[j] + "N"*(seqdiff)
            last = cimgt[-1]
            cimgt.extend([last]*(imgtdiff))
            clones[j].setField("imgtpartlabels",cimgt)

    if meta_data is not None:
        meta_data_ar = meta_data[0].split(",")
    for c in clones:
        if meta_data is not None:
            c.setField(meta_data[0],c.getField(meta_data_ar[0]))
            for m in range(1,len(meta_data_ar)):
                st = c.getField(meta_data[0])+":"+c.getField(meta_data_ar[m])
                c.setField(meta_data[0],st)

        if len(c.getField("imgtpartlabels")) != len(imgtar):
            printError("IMGT assignments are not the same within clone %d!\n" % c.clone,False)
            printError(c.getField("imgtpartlabels"),False)
            printError("%s\n%d\n" % (imgtar,j),False)
            for j in range(0, len(sequences)):
                printError("%s\n%s\n" % (sequences[j],clones[j].getField("imgtpartlabels")),False)
            printError("ChangeO file needs to be corrected")
        for j in range(0,len(imgtar)):
            if c.getField("imgtpartlabels")[j] != imgtar[j]:
                printError("IMGT assignments are not the same within clone %d!\n" % c.clone, False)
                printError(c.getField("imgtpartlabels"), False)
                printError("%s\n%d\n" % (imgtar, j))

    #Resolve germline if there are differences, e.g. if reconstruction was done before clonal clustering
    resolveglines = False
    for c in clones:
        ngermline = c.getField("germline_imgt_d_mask")
        if ngermline == "":
            ngermline = c.getField("germline_imgt")
        if ngermline != germline:
            resolveglines = True

    if resolveglines:
        printError("%s %s" % ("Predicted germlines are not the same among sequences in the same clone.",
                              "Be sure to cluster sequences into clones first and then predict germlines using --cloned"))

    if sites > (len(germline)):
        seqdiff = sites - len(germline)
        germline = germline + "N" * (seqdiff)

    if sites % 3 != 0:
        printError("number of sites must be divisible by 3! len: %d, clone: %s , id: %s, seq: %s" %(len(sequences[0]),\
        clones[0].clone,clones[0].sequence_id,sequences[0]))
    return imgtar, germline, sites, nseqs

def outputSeqPartFiles(out_dir, useqs_f, meta_data, clones, collapse, nseqs, delim, newgerm, conseqs, duplicate, imgt):
    """
    Create intermediate sequence alignment and partition files for IgPhyML output

    Arguments:
        out_dir (str): directory for sequence files.
        useqs_f (dict): unique sequences mapped to ids.
        meta_data (str): Field to append to sequence IDs. Splits identical sequences with different meta_data.
        clones (list) : list of receptor objects.
        collapse (bool) : deduplicate sequences.
        nseqs (int): number of sequences.
        delim (str) : delimiter for extracting metadata from ID.
        newgerm (str) : modified germline of clonal lineage.
        conseqs (list) : consensus sequences.
        duplicate (bool) : duplicate sequence if only one in a clone.
        imgt (list) : IMGT numbering of clonal positions .
    """

    # bootstrap these data if desired
    lg = len(newgerm)
    sites = range(0, lg)

    transtable = clones[0].sequence_id.maketrans(" ", "_")
    outfile = os.path.join(out_dir, "%s.fasta" % clones[0].clone)
    with open(outfile, "w") as clonef:
        if collapse:
            for seq_f, num in useqs_f.items():
                seq = seq_f
                cid = ""
                if meta_data is not None:
                    seq, cid = seq_f.split(delim)
                    cid = delim + cid.replace(":", "_")
                sid = clones[num].sequence_id.translate(transtable) + cid
                clonef.write(">%s\n%s\n" % (sid.replace(":","-"), seq.replace(".", "-")))
                if len(useqs_f) == 1 and duplicate:
                    if meta_data is not None:
                        if meta_data[0] == "DUPCOUNT":
                            cid = delim + "0"
                    sid = clones[num].sequence_id.translate(transtable) + "_1" + cid
                    clonef.write(">%s\n%s\n" % (sid.replace(":","-"), seq.replace(".", "-")))
        else:
            for j in range(0, nseqs):
                cid = ""
                if meta_data is not None:
                    meta_data_list = []
                    for m in meta_data:
                        meta_data_list.append(clones[j].getField(m).replace(":", "_"))
                    cid = delim + str(delim.join(meta_data_list))
                sid = clones[j].sequence_id.translate(transtable) + cid
                clonef.write(">%s\n%s\n" % (sid.replace(":","-"), conseqs[j].replace(".", "-")))
                if nseqs == 1 and duplicate:
                    if meta_data is not None:
                        if meta_data[0] == "DUPCOUNT":
                            cid = delim + "0"
                    sid = clones[j].sequence_id.translate(transtable)+"_1" + cid
                    clonef.write(">%s\n%s\n" % (sid.replace(":","-"), conseqs[j].replace(".", "-")))

        germ_id = ["GERM"]
        if meta_data is not None:
            for i in range(1,len(meta_data)):
                germ_id.append("GERM")
        clonef.write(">%s_%s\n" % (clones[0].clone,"_".join(germ_id)))
        for i in range(0, len(newgerm)):
            clonef.write("%s" % newgerm[i].replace(".","-"))
        clonef.write("\n")

    #output partition file
    partfile = os.path.join(out_dir, "%s.part.txt" % clones[0].clone)
    with open(partfile, "w") as partf:
        partf.write("%d %d\n" % (2, len(newgerm)))
        partf.write("FWR:IMGT\n")
        partf.write("CDR:IMGT\n")
        partf.write("%s\n" % (clones[0].v_call.split("*")[0]))
        partf.write("%s\n" % (clones[0].j_call.split("*")[0]))
        partf.write(",".join(map(str, imgt)))
        partf.write("\n")


def outputIgPhyML(clones, sequences, meta_data=None, collapse=False, ncdr3=False, logs=None,
                  fail_writer=None, out_dir=None, min_seq=1):
    """
    Create intermediate sequence alignment and partition files for IgPhyML output

    Arguments:
      clones (list): receptor objects within the same clone.
      sequences (list): sequences within the same clone (share indexes with clones parameter).
      meta_data (str): Field to append to sequence IDs. Splits identical sequences with different meta_data
      collapse (bool): if True collapse identical sequences.
      ncdr3 (bool): if True remove CDR3
      logs (dict): contains log information for each sequence
      out_dir (str): directory for output files.
      fail_writer (changeo.IO.TSVWriter): failed sequences writer object.
      min_seq (int): minimum number of data sequences to include.

    Returns:
      int: number of clones.
    """
    s = ""
    delim = "_"
    duplicate = True # duplicate sequences in clones with only 1 sequence?
    imgtar, germline, sites, nseqs = characterizePartitionErrors(sequences, clones, meta_data)

    tallies = []
    for i in range(0, sites, 3):
        tally = 0
        for j in range(0, nseqs):
            if sequences[j][i:(i + 3)] != "...":
                tally += 1
        tallies.append(tally)

    newseqs = []  # remove gap only sites from observed data
    newgerm = []
    imgt = []
    for j in range(0, nseqs):
        for i in range(0, sites, 3):
            if i == 0:
                newseqs.append([])
            if tallies[i//3] > 0:
                newseqs[j].append(sequences[j][i:(i+3)])
    lcodon = ""
    for i in range(0, sites, 3):
        if tallies[i//3] > 0:
            newgerm.append(germline[i:(i+3)])
            lcodon=germline[i:(i+3)]
            imgt.append(imgtar[i])

    if len(lcodon) == 2:
        newgerm[-1] = newgerm[-1] + "N"
    elif len(lcodon) == 1:
        newgerm[-1] = newgerm[-1] + "NN"

    if ncdr3:
        ngerm = []
        nimgt = []
        for i in range(0, len(newseqs)):
            nseq = []
            ncdr3 = 0
            for j in range(0, len(imgt)):
                if imgt[j] != 108:
                    nseq.append(newseqs[i][j])
                    if i == 0:
                        ngerm.append(newgerm[j])
                        nimgt.append(imgt[j])
                else:
                    ncdr3 += 1
            newseqs[i] = nseq
        newgerm = ngerm
        imgt = nimgt
            #print("Length: " + str(ncdr3))

    useqs_f = OrderedDict()
    conseqs = []
    for j in range(0, nseqs):
        conseq = "".join([str(seq_rec) for seq_rec in newseqs[j]])
        if meta_data is not None:
            meta_data_list = []
            for m in range(0,len(meta_data)):
                if isinstance(clones[j].getField(meta_data[m]), str):
                    clones[j].setField(meta_data[m],clones[j].getField(meta_data[m]).replace("_", ""))
                meta_data_list.append(str(clones[j].getField(meta_data[m])))
            conseq_f = "".join([str(seq_rec) for seq_rec in newseqs[j]])+delim+":".join(meta_data_list)
        else:
            conseq_f = conseq
        if conseq_f in useqs_f and collapse:
            clones[useqs_f[conseq_f]].dupcount += clones[j].dupcount
            logs[clones[j].sequence_id]["PASS"] = False
            logs[clones[j].sequence_id]["FAIL"] = "Duplication of " + clones[useqs_f[conseq_f]].sequence_id
            logs[clones[j].sequence_id]["DUPLICATE"]=True
            if fail_writer is not None:
                fail_writer.writeReceptor(clones[j])
        else:
            useqs_f[conseq_f] = j
        conseqs.append(conseq)

    if collapse:
        useqs_f = deduplicate(useqs_f, clones, logs, meta_data, delim)

    if collapse and len(useqs_f) < min_seq:
        for seq_f, num in useqs_f.items():
            logs[clones[num].sequence_id]["FAIL"] = "Clone too small: " + str(len(useqs_f))
            logs[clones[num].sequence_id]["PASS"] = False
        return -len(useqs_f)
    elif not collapse and len(conseqs) < min_seq:
        for j in range(0, nseqs):
            logs[clones[j].sequence_id]["FAIL"] = "Clone too small: " + str(len(conseqs))
            logs[clones[j].sequence_id]["PASS"] = False
        return -len(conseqs)

    # Output fasta file of masked, concatenated sequences
    outputSeqPartFiles(out_dir, useqs_f, meta_data, clones, collapse, nseqs,
                       delim, newgerm, conseqs, duplicate, imgt)

    if collapse:
        return len(useqs_f)
    else:
        return nseqs


def maskCodonsLoop(r, clones, cloneseqs, logs, fails, out_args, fail_writer, mask=True):
    """
    Masks codons split by alignment to IMGT reference

    Arguments:
      r (changeo.Receptor.Receptor): receptor object for a particular sequence.
      clones (list): list of receptors.
      cloneseqs (list): list of masked clone sequences.
      logs (dict): contains log information for each sequence.
      fails (dict): counts of various sequence processing failures.
      out_args (dict): arguments for output preferences.
      fail_writer (changeo.IO.TSVWriter): failed sequences writer object.

    Returns:
      0: returns 0 if an error occurs or masking fails.
      1: returns 1 masking succeeds
    """
    if r.clone is None:
        printError("Cannot export datasets until sequences are clustered into clones.")
    if r.dupcount is None:
        r.dupcount = 1
    fails["rec_count"] += 1
    #printProgress(rec_count, rec_count, 0.05, start_time)
    ptcs = hasPTC(r.sequence_imgt)
    gptcs = hasPTC(r.getField("germline_imgt_d_mask"))
    if gptcs >= 0:
        log = OrderedDict()
        log["ID"] = r.sequence_id
        log["CLONE"] = r.clone
        log["SEQ_IN"] = r.sequence_input
        log["SEQ_IMGT"] = r.sequence_imgt
        logs[r.sequence_id] = log
        logs[r.sequence_id]["PASS"] = False
        logs[r.sequence_id]["FAIL"] = "Germline PTC"
        fails["seq_fail"] += 1
        fails["germlineptc"] += 1
        return 0

    if r.functional and ptcs < 0:
        #If IMGT regions are provided, record their positions
        rd = RegionDefinition(r.junction_length, amino_acid=False)
        regions = rd.getRegions(r.sequence_imgt)
        if regions["cdr3_imgt"] != "" and regions["cdr3_imgt"] is not None:
            simgt = regions["fwr1_imgt"] + regions["cdr1_imgt"] + regions["fwr2_imgt"] + regions["cdr2_imgt"] + \
                    regions["fwr3_imgt"] + regions["cdr3_imgt"] + regions["fwr4_imgt"]
            if len(simgt) < len(r.sequence_imgt):
                r.fwr4_imgt = r.fwr4_imgt +  ("."*(len(r.sequence_imgt) - len(simgt)))
                simgt = regions["fwr1_imgt"] + regions["cdr1_imgt"] + regions["fwr2_imgt"] + \
                        regions["cdr2_imgt"] + regions["fwr3_imgt"] + regions["cdr3_imgt"] + regions["fwr4_imgt"]
            imgtpartlabels = [13]*len(regions["fwr1_imgt"]) + [30]*len(regions["cdr1_imgt"]) + [45]*len(regions["fwr2_imgt"]) + \
                             [60]*len(regions["cdr2_imgt"]) + [80]*len(regions["fwr3_imgt"]) + [108] * len(regions["cdr3_imgt"]) + \
                             [120] * len(regions["fwr4_imgt"])
            r.setField("imgtpartlabels", imgtpartlabels)

            if len(r.getField("imgtpartlabels")) != len(r.sequence_imgt) or simgt != r.sequence_imgt:
                log = OrderedDict()
                log["ID"] = r.sequence_id
                log["CLONE"] = r.clone
                log["SEQ_IN"] = r.sequence_input
                log["SEQ_IMGT"] = r.sequence_imgt
                logs[r.sequence_id] = log
                logs[r.sequence_id]["PASS"] = False
                logs[r.sequence_id]["FAIL"] = "FWR/CDR error"
                logs[r.sequence_id]["FWRCDRSEQ"] = simgt
                fails["seq_fail"] += 1
                fails["region_fail"] += 1
                return 0

        elif regions["fwr3_imgt"] != "" and regions["fwr3_imgt"] is not None:
            simgt = regions["fwr1_imgt"] + regions["cdr1_imgt"] + regions["fwr2_imgt"] + regions["cdr2_imgt"] + \
                    regions["fwr3_imgt"]
            nseq = r.sequence_imgt[len(simgt):len(r.sequence_imgt)]

            if len(simgt) < len(r.sequence_imgt):
                simgt = regions["fwr1_imgt"] + regions["cdr1_imgt"] + regions["fwr2_imgt"] + \
                        regions["cdr2_imgt"] + regions["fwr3_imgt"] + nseq
            imgtpartlabels = [13] * len(regions["fwr1_imgt"]) + [30] * len(regions["cdr1_imgt"]) + [45] * len(
                regions["fwr2_imgt"]) + \
                             [60] * len(regions["cdr2_imgt"]) + [80] * len(regions["fwr3_imgt"]) + \
                             [108] * int(len(nseq))
            r.setField("imgtpartlabels", imgtpartlabels)
            if len(r.getField("imgtpartlabels")) != len(r.sequence_imgt) or simgt != r.sequence_imgt:
                log = OrderedDict()
                log["ID"] = r.sequence_id
                log["CLONE"] = r.clone
                log["SEQ_IN"] = r.sequence_input
                log["SEQ_IMGT"] = r.sequence_imgt
                logs[r.sequence_id] = log
                logs[r.sequence_id]["PASS"] = False
                logs[r.sequence_id]["FAIL"] = "FWR/CDR error"
                logs[r.sequence_id]["FWRCDRSEQ"] = simgt
                fails["seq_fail"] += 1
                fails["region_fail"] += 1
                return 0
        else:
            #imgt_warn = "\n! IMGT FWR/CDR sequence columns not detected.\n! Cannot run CDR/FWR partitioned model on this data.\n"
            imgtpartlabels = [0] * len(r.sequence_imgt)
            r.setField("imgtpartlabels", imgtpartlabels)
        mout = maskSplitCodons(r, mask=mask)
        mask_seq = mout[0]
        ptcs = hasPTC(mask_seq)
        if ptcs >= 0:
            printWarning("Masked sequence suddenly has a PTC.. %s\n" % r.sequence_id)
            mout[1]["PASS"] = False
            mout[1]["FAIL"] = "PTC_ADDED_FROM_MASKING"
        logs[mout[1]["ID"]] = mout[1]
        if mout[1]["PASS"]:
            #passreads += r.dupcount
            if r.clone in clones:
                clones[r.clone].append(r)
                cloneseqs[r.clone].append(mask_seq)
            else:
                clones[r.clone] = [r]
                cloneseqs[r.clone] = [mask_seq]
            return 1
        else:
            if out_args["failed"]:
                fail_writer.writeReceptor(r)
            fails["seq_fail"] += 1
            fails["failreads"] += r.dupcount
            if mout[1]["FAIL"] == "FRAME-SHIFTING DELETION":
                fails["del_fail"] += 1
            elif mout[1]["FAIL"] == "SINGLE FRAME-SHIFTING INSERTION":
                fails["in_fail"] += 1
            else:
                fails["other_fail"] += 1
    else:
        log = OrderedDict()
        log["ID"] = r.sequence_id
        log["CLONE"] = r.clone
        log["PASS"] = False
        log["FAIL"] = "NONFUNCTIONAL/PTC"
        log["SEQ_IN"] = r.sequence_input
        logs[r.sequence_id] = log
        if out_args["failed"]:
            fail_writer.writeReceptor(r)
        fails["seq_fail"] += 1
        fails["nf_fail"] += 1

    return 0

# Run IgPhyML on outputed data
def runIgPhyML(outfile, igphyml_out, clone_dir, nproc=1, optimization="lr", omega="e,e", kappa="e", motifs="FCH",
               hotness="e,e,e,e,e,e", oformat="tab", nohlp=False, asr=-1, clean="none"):
    """
    Run IgPhyML on outputted data

    Arguments:
      outfile (str): Output file name.
      igphymlout (str): igphyml output file
      nproc (int): Number of threads to parallelize IgPhyML across
      optimization (str): Optimize combination of topology (t) branch lengths (l) and parameters (r) in IgPhyML.
      omega (str): omega optimization in IgPhyML (--omega)
      kappa (str): kappa optimization in IgPhyML (-t)
      motifs (str): motifs to use in IgPhyML (--motifs)
      hotness (str): motif in IgPhyML (--hotness)
      oformat (str): output format for IgPhyML (tab or txt)
      nohlp (bool): If True, only estimate GY94 trees and parameters
      clean (str): delete intermediate files? (none, all)
    """
    osplit = outfile.split(".")
    outrep = ".".join(osplit[0:(len(osplit)-1)]) + "_gy.tsv"
    gyout = outfile + "_igphyml_stats_gy.txt"

    gy_args = ["igphyml", "--repfile", outfile, "-m", "GY", "--run_id", "gy", "--outrep", outrep, "--threads",
               str(nproc),"--outname",gyout]


    hlp_args = ["igphyml","--repfile", outrep, "-m", "HLP", "--run_id", "hlp", "--threads", str(nproc), "-o",
                optimization, "--omega", omega, "-t", kappa, "--motifs", motifs, "--hotness", hotness, "--oformat",
                oformat, "--outname", igphyml_out]

    if asr >= 0:
        hlp_args.append("--ASRc")
        hlp_args.append(str(asr))

    log = OrderedDict()
    log["START"] = "IgPhyML GY94 tree estimation"
    printLog(log)

    try: #check for igphyml executable
        subprocess.check_output(["igphyml"])
    except:
        printError("igphyml not found :-/")
    try: #get GY94 starting topologies
        p = subprocess.check_output(gy_args)
    except subprocess.CalledProcessError as e:
        print(" ".join(gy_args))
        print('error>', e.output, '<')
        printError("GY94 tree building in IgPhyML failed")

    log = OrderedDict()
    log["START"] = "IgPhyML HLP analysis"
    log["OPTIMIZE"] = optimization
    log["TS/TV"] = kappa
    log["wFWR,wCDR"] = omega
    log["MOTIFS"] = motifs
    log["HOTNESS"] = hotness
    log["NPROC"] = nproc
    printLog(log)

    if not nohlp:
        try: #estimate HLP parameters/trees
            p = subprocess.check_output(hlp_args)
        except subprocess.CalledProcessError as e:
            print(" ".join(hlp_args))
            print('error>', e.output, '<')
            printError("HLP tree building failed")

    log = OrderedDict()
    log["OUTPUT"] = igphyml_out
    if oformat == "tab":
        igf = open(igphyml_out)
        names = igf.readline().split("\t")
        vals = igf.readline().split("\t")
        for i in range(3,len(names)-1):
            log[names[i]] = round(float(vals[i]),2)
    printLog(log)

    if clean != "none":
        log = OrderedDict()
        log["START"] = "CLEANING"
        log["SCOPE"] = clean
        printLog(log)
        todelete = open(outrep)
        for line in todelete:
            line = line.rstrip("\n")
            line = line.rstrip("\r")
            lsplit = line.split("\t")
            if len(lsplit) == 4:
                os.remove(lsplit[0])
                os.remove(lsplit[1])
                os.remove(lsplit[3])
        todelete.close()
        os.remove(outrep)
        os.remove(outfile)
        os.remove(gyout)
        cilog = outrep + "_igphyml_CIlog.txt_hlp"
        if os.path.isfile(cilog):
            os.remove(cilog)
        if oformat == "tab":
            os.rmdir(clone_dir)
        else:
            printWarning("Using --clean all with --oformat txt will not delete all tree file results.\n"
                         "You'll have to do that yourself.")
        log = OrderedDict()
        log["END"] = "IgPhyML analysis"
        printLog(log)


# Note: Collapse can give misleading dupcount information if some sequences have ambiguous characters at polymorphic sites
def buildTrees(db_file, meta_data=None, target_clones=None, collapse=False, ncdr3=False, nmask=False,
               sample_depth=-1, min_seq=1,append=None, igphyml=False, nproc=1, optimization="lr", omega="e,e",
               kappa="e", motifs="FCH", hotness="e,e,e,e,e,e", oformat="tab", clean="none", nohlp=False,
               asr=-1, format=default_format, out_args=default_out_args):
    """
    Masks codons split by alignment to IMGT reference, then produces input files for IgPhyML

    Arguments:
      db_file (str): input tab-delimited database file.
      meta_data (str): Field to append to sequence IDs. Splits identical sequences with different meta_data
      target_clones (str): List of clone IDs to analyze.
      collapse (bool): if True collapse identical sequences.
      ncdr3 (bool): if True remove all CDR3s.
      nmask (bool): if False, do not attempt to mask split codons
      sample_depth (int): depth of subsampling before deduplication
      min_seq (int): minimum number of sequences per clone
      append (str): column name to append to sequence_id
      igphyml (bool): If True, run IgPhyML on outputted data
      nproc (int) : Number of threads to parallelize IgPhyML across
      optimization (str): Optimize combination of topology (t) branch lengths (l) and parameters (r) in IgPhyML.
      omega (str): omega optimization in IgPhyML (--omega)
      kappa (str): kappa optimization in IgPhyML (-t)
      motifs (str): motifs to use in IgPhyML (--motifs)
      hotness (str): motif in IgPhyML (--hotness)
      oformat (str): output format for IgPhyML (tab or txt)
      clean (str): delete intermediate files? (none, all)
      nohlp (bool): If True, only estimate GY94 trees and parameters
      format (str): input and output format.
      out_args (dict): arguments for output preferences.

    Returns:
      dict: dictionary of output pass and fail files.
    """
    # Print parameter info
    log = OrderedDict()
    log["START"] = "BuildTrees"
    log["FILE"] = os.path.basename(db_file)
    log["COLLAPSE"] = collapse
    printLog(log)

    # Open output files
    out_label = "lineages"

    pass_handle = getOutputHandle(db_file,
                                  out_label=out_label,
                                  out_dir=out_args["out_dir"],
                                  out_name= out_args["out_name"],
                                  out_type="tsv")

    igphyml_out = None
    if igphyml:
        igphyml_out = getOutputName(db_file, out_label="igphyml-pass",
                                    out_dir=out_args["out_dir"],
                                    out_name=out_args["out_name"],
                                    out_type=oformat)

    dir_name, __ = os.path.split(pass_handle.name)

    if out_args["out_name"] is None:
        __, clone_name, __ = splitName(db_file)
    else:
        clone_name = out_args["out_name"]
    if dir_name is None:
        clone_dir = clone_name
    else:
        clone_dir = os.path.join(dir_name, clone_name)
    if not os.path.exists(clone_dir):
        os.makedirs(clone_dir)

    # Format options
    try:
        reader, writer, __ = getFormatOperators(format)
    except ValueError:
        printError("Invalid format %s." % format)
    out_fields = getDbFields(db_file, reader=reader)

    # open input file
    handle = open(db_file, "r")
    records = reader(handle)

    fail_handle, fail_writer = None, None
    if out_args["failed"]:
        fail_handle = getOutputHandle(db_file,
                                      out_label="lineages-fail",
                                      out_dir=out_args["out_dir"],
                                      out_name=out_args["out_name"],
                                      out_type=out_args["out_type"])
        fail_writer = writer(fail_handle, fields=out_fields)

    cloneseqs = {}
    clones = {}
    logs = OrderedDict()
    fails = {"rec_count":0, "seq_fail":0, "nf_fail":0, "del_fail":0, "in_fail":0, "minseq_fail":0,
             "other_fail":0, "region_fail":0, "germlineptc":0, "fdcount":0, "totalreads":0,
             "passreads":0, "failreads":0}

    # Mask codons split by indels
    start_time = time()
    printMessage("Correcting frames and indels of sequences", start_time=start_time, width=50)

    #subsampling loop
    init_clone_sizes = {}
    big_enough = []
    all_records = []
    found_no_funct = False
    for r in records:
        if r.functional is None:
            r.functional = True
            if found_no_funct is False:
                printWarning("FUNCTIONAL column not found.")
                found_no_funct = True
        all_records.append(r)
        if r.clone in init_clone_sizes:
            init_clone_sizes[r.clone] += 1
        else:
            init_clone_sizes[r.clone] = 1

    for r in all_records:
        if target_clones is None or r.clone in target_clones:
            if init_clone_sizes[r.clone] >= min_seq:
               big_enough.append(r)

    fails["totalreads"] = len(all_records)
    #fails["minseq_fail"] =  len(all_records) - len(big_enough)         

    if len(big_enough) == 0:
        printError("\n\nNo sequences found that match specified criteria.",1)

    if sample_depth > 0:
        random.shuffle(big_enough)

    total = 0
    for r in big_enough:
        if r.functional is None:
            r.functional = True
            if found_no_funct is False:
                printWarning("FUNCTIONAL column not found.")
                found_no_funct = True

        r.sequence_id = r.sequence_id.replace(",","-") #remove commas from sequence ID
        r.sequence_id = r.sequence_id.replace(":","-") #remove colons from sequence ID
        r.sequence_id = r.sequence_id.replace(",","-") #remove commas from sequence ID
        r.sequence_id = r.sequence_id.replace(")","-") #remove parenthesis from sequence ID
        r.sequence_id = r.sequence_id.replace("(","-") #remove parenthesis from sequence ID
        if(meta_data is not None):
            for m in range(0,len(meta_data)):
                md = r.getField(meta_data[m])
                md = md.replace(",","-") #remove commas from metadata
                md = md.replace(":","-") #remove colons from metadata
                md = md.replace(",","-") #remove commas from metadata
                md = md.replace(")","-") #remove parenthesis from metadata
                md = md.replace("(","-") #remove parenthesis from metadata
                r.setField(meta_data[m],md)

        if append is not None:
            if append is not None:
                for m in append:
                    r.sequence_id = r.sequence_id + "_" + r.getField(m)
        total += maskCodonsLoop(r, clones, cloneseqs, logs, fails, out_args, fail_writer,
                                mask = not nmask)
        if total == sample_depth:
            break

    # Start processing clones
    clonesizes = {}
    pass_count, nclones = 0, 0
    printMessage("Processing clones", start_time=start_time, width=50)
    for k in clones.keys():
        if len(clones[str(k)]) < min_seq:
            for j in range(0, len(clones[str(k)])):
                logs[clones[str(k)][j].sequence_id]["FAIL"] = "Clone too small: " + str(len(cloneseqs[str(k)]))
                logs[clones[str(k)][j].sequence_id]["PASS"] = False
            clonesizes[str(k)] = -len(cloneseqs[str(k)])
        else:
            clonesizes[str(k)] = outputIgPhyML(clones[str(k)], cloneseqs[str(k)], meta_data=meta_data, collapse=collapse,
                                           ncdr3=ncdr3, logs=logs, fail_writer=fail_writer,
                                           out_dir=clone_dir, min_seq=min_seq)

        #If clone is too small, size is returned as a negative
        if clonesizes[str(k)] > 0:
            nclones += 1
            pass_count += clonesizes[str(k)]
        else:
            fails["seq_fail"] -= clonesizes[str(k)]
            fails["minseq_fail"]  -= clonesizes[str(k)]
    fail_count = fails["rec_count"] - pass_count

    # End clone processing
    printMessage("Done", start_time=start_time, end=True, width=50)

    log_handle = None
    if out_args["log_file"] is not None:
        log_handle = open(out_args["log_file"], "w")
        for j in logs.keys():
            printLog(logs[j], handle=log_handle)

    pass_handle.write(str(nclones)+"\n")
    for key in sorted(clonesizes, key=clonesizes.get, reverse=True):
        #print(key + "\t" + str(clonesizes[key]))
        outfile = os.path.join(clone_dir, "%s.fasta" % key)
        partfile = os.path.join(clone_dir, "%s.part.txt" % key)
        if clonesizes[key] > 0:
            germ_id = ["GERM"]
            if meta_data is not None:
                for i in range(1, len(meta_data)):
                    germ_id.append("GERM")
            pass_handle.write("%s\t%s\t%s_%s\t%s\n" % (outfile, "N", key,"_".join(germ_id), partfile))

    handle.close()
    output = {"pass": None, "fail": None}
    if pass_handle is not None:
        output["pass"] = pass_handle.name
        pass_handle.close()
    if fail_handle is not None:
        output["fail"] = fail_handle.name
        fail_handle.close()
    if log_handle is not None:
        log_handle.close()

    #printProgress(rec_count, rec_count, 0.05, start_time)
    log = OrderedDict()
    log["OUTPUT"] = os.path.basename(pass_handle.name) if pass_handle is not None else None
    log["RECORDS"] = fails["totalreads"]
    log["INITIAL_FILTER"] = fails["rec_count"]
    log["PASS"] = pass_count
    log["FAIL"] = fail_count
    log["NONFUNCTIONAL"] = fails["nf_fail"]
    log["FRAMESHIFT_DEL"] = fails["del_fail"]
    log["FRAMESHIFT_INS"] = fails["in_fail"]
    log["CLONETOOSMALL"] = fails["minseq_fail"]
    log["CDRFWR_ERROR"] = fails["region_fail"]
    log["GERMLINE_PTC"] = fails["germlineptc"]
    log["OTHER_FAIL"] = fails["other_fail"]

    if collapse:
        log["DUPLICATE"] = fail_count - fails["seq_fail"]
    log["END"] = "BuildTrees"
    printLog(log)

    #Run IgPhyML on outputted data?
    if igphyml:
        runIgPhyML(pass_handle.name, igphyml_out=igphyml_out, clone_dir=clone_dir, nproc=nproc,
                   optimization=optimization, omega=omega, kappa=kappa, motifs=motifs,
                   hotness=hotness, oformat=oformat, nohlp=nohlp,clean=clean,asr=asr)

    return output


def getArgParser():
    """
    Defines the ArgumentParser

    Returns:
        argparse.ArgumentParser: argument parsers.
    """
    # Define input and output field help message
    fields = dedent(
             """
             output files:
                 <folder>
                   folder containing fasta and partition files for each clone.
                 lineages
                    successfully processed records.
                 lineages-fail
                    database records failed processing.
                 igphyml-pass
                    parameter estimates and lineage trees from running IgPhyML, if specified

             required fields:
                 sequence_id, sequence, sequence_alignment,
                 germline_alignment_d_mask or germline_alignment,
                 v_call, j_call, clone_id, v_sequence_start
              """)

    # Parent parser
    parser_parent = getCommonArgParser(out_file=False, log=True, format=True)

    # Define argument parser
    parser = ArgumentParser(description=__doc__, epilog=fields,
                            parents=[parser_parent],
                            formatter_class=CommonHelpFormatter, add_help=False)

    group = parser.add_argument_group("sequence processing arguments")
    group.add_argument("--collapse", action="store_true", dest="collapse",
                        help="""If specified, collapse identical sequences before exporting to fasta.""")
    group.add_argument("--ncdr3", action="store_true", dest="ncdr3",
                        help="""If specified, remove CDR3 from all sequences.""")
    group.add_argument("--nmask", action="store_true", dest="nmask",
                       help="""If specified, do not attempt to mask split codons.""")
    group.add_argument("--md", nargs="+", action="store", dest="meta_data",
                       help="""List of fields to containing metadata to include in output fasta file 
                            sequence headers.""")
    group.add_argument("--clones", nargs="+", action="store", dest="target_clones",
                       help="""List of clone IDs to output, if specified.""")
    group.add_argument("--minseq", action="store", dest="min_seq", type=int, default=1,
                       help="""Minimum number of data sequences. Any clones with fewer than the specified
                            number of sequences will be excluded.""")
    group.add_argument("--sample", action="store", dest="sample_depth", type=int, default=-1,
                       help="""Depth of reads to be subsampled (before deduplication).""")
    group.add_argument("--append", nargs="+", action="store", dest="append",
                       help="""List of columns to append to sequence ID to ensure uniqueness.""")

    igphyml_group = parser.add_argument_group("IgPhyML arguments (see igphyml -h for details)")
    igphyml_group.add_argument("--igphyml", action="store_true", dest="igphyml",
                               help="""Run IgPhyML on output?""")
    igphyml_group.add_argument("--nproc", action="store", dest="nproc", type=int, default=1,
                               help="""Number of threads to parallelize IgPhyML across.""")
    igphyml_group.add_argument("--clean", action="store", choices=("none", "all"),
                               dest="clean", type=str, default="none",
                               help="""Delete intermediate files? 
                               none: leave all intermediate files; all: delete all intermediate files.""")
    igphyml_group.add_argument("--optimize", action="store", dest="optimization", type=str, default="lr",
                               choices=("n","r","l","lr","tl","tlr"),
                               help="""Optimize combination of topology (t) branch lengths (l) and parameters (r), or 
                               nothing (n), for IgPhyML.""")
    igphyml_group.add_argument("--omega", action="store", dest="omega", type=str, default="e,e",
                               help="""Omega parameters to estimate for FWR,CDR respectively: 
                               e = estimate, ce = estimate + confidence interval, or numeric value""")
    igphyml_group.add_argument("-t", action="store", dest="kappa", type=str, default="e",
                               help="""Kappa parameters to estimate: 
                               e = estimate, ce = estimate + confidence interval, or numeric value""")
    igphyml_group.add_argument("--motifs", action="store", dest="motifs", type=str,
                               default="WRC_2:0,GYW_0:1,WA_1:2,TW_0:3,SYC_2:4,GRS_0:5",
                               help="""Which motifs to estimate mutability.""")
    igphyml_group.add_argument("--hotness", action="store", dest="hotness", type=str, default="e,e,e,e,e,e",
                               help="""Mutability parameters to estimate: 
                               e = estimate, ce = estimate + confidence interval, or numeric value""")
    igphyml_group.add_argument("--oformat", action="store", dest="oformat", type=str, default="tab",
                               choices=("tab", "txt"),
                               help="""IgPhyML output format.""")
    igphyml_group.add_argument("--nohlp", action="store_true", dest="nohlp",
                               help="""Don't run HLP model?""")
    igphyml_group.add_argument("--asr", action="store", dest="asr", type=float, default=-1,
                               help="""Ancestral sequence reconstruction interval (0-1).""")

    return parser


if __name__ == "__main__":
    """
    Parses command line arguments and calls main
    """
    # Parse command line arguments
    parser = getArgParser()
    checkArgs(parser)
    args = parser.parse_args()
    args_dict = parseCommonArgs(args)
    del args_dict["db_files"]

    # Call main for each input file
    for f in args.__dict__["db_files"]:
        args_dict["db_file"] = f
        buildTrees(**args_dict)


"""
Default parameters
"""

# Info
__author__ = 'Jason Anthony Vander Heiden, Namita Gupta'

# System settings
default_csv_size = 2**24

# Fields
default_v_field = 'v_call'
default_d_field = 'd_call'
default_j_field = 'j_call'
default_id_field = 'sequence_id'
default_seq_field = 'sequence_alignment'
default_germ_field = 'germline_alignment'
default_junction_field = 'junction'
default_clone_field = 'clone_id'

# Receptor attributes
v_attr = 'v_call'
d_attr = 'd_call'
j_attr = 'j_call'
id_attr = 'sequence_id'
seq_attr = 'sequence_imgt'
germ_attr = 'germline_imgt'
junction_attr = 'junction'
clone_attr = 'clone'

# External applications
default_igblastn_exec = 'igblastn'
default_igblastp_exec = 'igblastp'
default_tbl2asn_exec = 'tbl2asn'
default_igphyml_exec = 'igphyml'

# Commandline arguments
choices_format = ('airr', 'changeo')
default_format = 'airr'
default_out_args = {'log_file': None,
                    'out_dir': None,
                    'out_name': None,
                    'out_type': 'tsv',
                    'failed': False}

# IMGT
default_imgt_id_len = 49
